# -*- coding: utf-8 -*-
# Copyright 2022 Kristo Mela
# This source code is licensed under the MIT license. See LICENSE in the repository root directory.
# Author(s): Kristo Mela
"""
Created on Tue Dec 28 13:46:06 2021

Class for Frame Members, to be used with Raami

@author: kmela
"""

import numpy as np
import matplotlib.pyplot as plt
import matplotlib.patches as patches
import matplotlib.lines as mlines
import matplotlib.path as mpath

from metku.structures.steel.steel_member import SteelMember
from metku.raami.frame_node import FrameNode
import metku.framefem.framefem
from metku.framefem.elements.ebbeam import EBBeam, EBBeam3D
from metku.framefem.elements.rod import Rod
from metku.sections.steel.catalogue import make_section, ipe_profiles, h_profiles, rhs_profiles, shs_profiles, chs_profiles, hea_profiles, heb_profiles
from metku.sections.steel.CHS import CHS
from metku.sections.steel.RHS import RHS, SHS
from metku.sections.steel.ISection import IPE, HEA, HEB

#from loadIDs import LoadIDs

from .frame_loads import LoadIDs, LineLoad

def profile_list(cross_section,prof_type="CURRENT",Tmin=3.0,sec_class=3):
    """ Returns a list of profiles matching the type of 'cross_section.
        This is used in 'optimum_design' methods.    
       
        Tmin .. minimum wall thickness, for RHS/SHS profiles
        sec_class .. maximum section class (in compression)
    """
    
    prof_type = prof_type.upper()
    
    if prof_type == "CURRENT":
        """ Use the cross section family of the current profile """
        if isinstance(cross_section,IPE):
            prof_type = "IPE"
        elif isinstance(cross_section,HEA):
            prof_type = "HEA"
        elif isinstance(cross_section,HEB):
            prof_type = "HEB"
        elif isinstance(cross_section,RHS):
            prof_type = "RHS"
        elif isinstance(cross_section,SHS):
            prof_type = "SHS"
        elif isinstance(cross_section,CHS):
            prof_type = "CHS"
        else:
            prof_type = "IPE"
    
    if prof_type == "IPE":            
        profiles = ipe_profiles.keys()
    elif prof_type == "H":
        profiles = h_profiles.keys()
    elif prof_type == "HEA":
        profiles = hea_profiles.keys()
    elif prof_type == "HEB":
        profiles = heb_profiles.keys()
    elif prof_type == "RHS":
        profiles = []
        for key, value in rhs_profiles.items():
            pnew = make_section(key)
            pnew.Ned = -100e3
            if pnew.section_class() > sec_class:
                continue
            if value['t'] >= Tmin:
                profiles.append(key)
        #profiles = rhs_profiles.keys()
    elif prof_type == "CHS":
        profiles = chs_profiles.keys()
    elif prof_type == "SHS":
        profiles = shs_profiles.keys()
    else:
        raise ValueError(f'"{prof_type}" is not a valid profile type!')
    
    return profiles


class FrameMember:
    """ General frame member class
        
        Relevant member data includes:
            1) Cross-section and material
            2) Nodes
            3) Type: beam, column, brace, etc.
    
        Main operations:
            - Use member data to generate FEM model
            - From FEM results, read internal forces and displacements
            - Design member based on internal forces and displacements
        
    """

    def __init__(self,nodes,section,mem_type="beam",mem_id="",nel=2,\
                 hinges=[False,False],lcr=[1.0,1.0],sway=False):
        """
        Constructor

        Parameters
        ----------
        nodes : list of FrameNode objects
            End nodes of the member.
        section : varies
            Cross-sectional data.
        mem_type: string
            Type of member: 'beam', 'column', 'brace', etc.
        nel: int
            Number of elements for dividing the member
        hinges: list (boolean)
            States whether or not the ends of the member are hinged.
        Returns
        -------
        None.

        """
        
        self.active = True
        self.resistance_check = True
        self.allowable_utilization = 1.0
        
        self.fem_nodes = []
        self.fem_elements = []
        
        self.member = None
        
        # Buckling length factor
        self.__lcr = None
        self.lcr = lcr
                
        self.nodes = nodes
        
        for node in nodes:
            node.members.append(self)
        
        self.symmetry_pair = None # Potential symmetry member pair
        self.group = [] # List of members that are in the same group. These have
                          # have the same cross-section as the current member.
        
        self.cross_section = section
        self.mtype = mem_type
        self.mem_id = mem_id
        self.frame = None
        self.num_elements = nel
        self.local_node_coords = [] # Local coordinates of member nodes in FEM model
        self.global_node_coords = [] # Local coordinates of member nodes in FEM model
        
        self.loads = [] # List of loads acting on the member
        
        self.nodal_forces = {} # Nodal forces in different load cases
        self.NEd = {} # Design value of the axial force (maximum) for each load case
        self.VzEd = {} # Design value of the shear force (maximum) for each load case
        self.MyEd = {} # Design value of the bending moment (maximum) for each load case
        self.utilization = {} # Utilization ratio for different load cases
        
        # Create nodal coordinates
        self.calc_nodal_coordinates(nel)
        
        # Flags for stating whether or not the member ends are hinged.
        
        self.hinges = hinges
        
        # Flag stating whether or not the member can buckle in sway mode
        self.sway = sway
        
        self.costs = {}
        
        self.self_weight = False
        
        
        
    def __repr__(self):        
        return f"{type(self).__name__} [{self.mem_id}]"
    
    @property
    def cross_section(self):        
        return self.__cross_section
    
    @cross_section.setter
    def cross_section(self,val):
        
        self.__cross_section = val
        if not self.member is None:
            self.member.profile = val            
        
        self.update_element_sections()
        
        if not self.symmetry_pair is None:
            self.symmetry_pair.__cross_section = val
            if not self.symmetry_pair.member is None:
                self.symmetry_pair.member.profile = val
                self.symmetry_pair.update_element_sections()
            
    @property
    def material(self):
        """ Material of the member """
        return self.cross_section.material
    
    @material.setter
    def material(self,val):
        """ Set member material """
        self.cross_section.material = val
    
    @property
    def lcr(self):
        return self.__lcr
    
    @lcr.setter
    def lcr(self,val):
        self.__lcr = val
        
        if not self.member is None:
            self.member.lcr = val
    
    def coords(self):
        """ Return nodal coordinates """
        x1, x2 = np.asarray([self.nodes[0].coords,self.nodes[1].coords])
        return x1, x2
    
    def length(self):
        """ Calculate the length of the member """
        start_node, end_node = self.coords()
        return np.linalg.norm(end_node - start_node)
    
    def weight(self):
        """
        Function that calculates member's weight
        :return weight: float, weight of the member
        """

        return self.cross_section.A * self.length() * self.cross_section.density
    
    def cost(self):
        """ Total cost of the member 
            By default this is the material cost
        
        """
        
        if not self.member is None:
            c = self.member.cost()
        else:
            c = self.length()*self.cross_section.cost()
        
        return c
    
    @property
    def dir_vector(self):
        """
        Returns direction vector of the member.
        :return:
        """
        start, end = self.coords()
        unit = np.array(end - start) / self.length()
        return unit
    
    def local_to_global(self, loc):
        """
        Returns local coordinate in global coordinates
        :param loc: local coordinate
        :return: global coordinate
        """
        Lx = self.length() * loc
        return self.nodes[0].coords + Lx * self.dir_vector

    @property
    def perpendicular(self):
        """
        Returns vector perpendicular to member.
        NOTE! This needs to be modified for 3D!        
        :return:
        """
        unit = self.dir_vector
        #if self.reverse:
        #    return np.array([-unit[1], unit[0]])

        return np.array([unit[1], -unit[0]])
    
    @property
    def angle(self):
        """
        Returns member's angle with respect to x-axis in radians
        """
        start_node, end_node = self.coords()
        x0, y0 = start_node
        x1, y1 = end_node
        if abs(x1 - x0) < 1e-6:
            # 90 degrees
            angle = 0.5*np.pi # np.radians(90)
        else:
            angle = np.arctan((y1 - y0) / (x1 - x0))
            if angle < 0:
                return np.pi + angle #np.radians(180) + angle
        return angle
    
    def add_self_weight(self):
        """ Adds self-weight to the member's loads
        """
        if not self.self_weight:
            self.self_weight = True
            load_id = LoadIDs['SELF-WEIGHT']
            
            # Default units: N/mm
            value = -self.cross_section.self_weight()
            self.frame.add(LineLoad(self, [value, value], 'y',load_id,ltype='dead',
                                    name='self-weight'))

    def remove_self_weight(self):
        """ Removes self-weight from loads
        """
        if self.self_weight:
            self.self_weight = False
        
    
    def calc_nodal_coordinates(self, num_elements=0):
        """ Calculates nodal coordinates along the member 
            
        """
                        
        self.global_node_coords = []
        
        if num_elements > 0:
            self.num_elements = num_elements
        
        self.local_node_coords = np.linspace(0, 1, self.num_elements + 1)
        #for loc in np.arange(0, 1 + dloc, dloc):
        for loc in self.local_node_coords:
            coord = self.local_to_global(loc)
            self.global_node_coords.append(coord)
    
    def calc_nodal_forces(self,load_id=LoadIDs['ULS']):
        """ Collects the internal forces at the nodes into
            'nodal_forces' attribute which is a dict.
        """
                
        def absmax(vals):
            min_val = min(vals)
            max_val = max(vals)
            abs_max = max(abs(min_val), abs(max_val))
            if abs_max == abs(min_val):
                return min_val
            else:
                return max_val

        """ Store also maximum bending moment, shear force and axial force """
        max_med = 0
        max_ved = 0
        max_ned = 0
        
        # set up the nodal_forces dict for this load case
        self.nodal_forces[load_id] = {'N':[],'Vz':[],'My':[]}
        
        for element in self.fem_elements:
            # The elements in fem_elements appear in the order from the first
            # end node to the last. Reading the forces in the first node of each element
            # generates a list of nodal forces that appear in the same order as
            # the local coordinates.
            self.nodal_forces[load_id]['N'].append(element.fint[load_id]['fx'][0])
            self.nodal_forces[load_id]['Vz'].append(element.fint[load_id]['fz'][0])
            self.nodal_forces[load_id]['My'].append(element.fint[load_id]['my'][0])

        # Read the forces in the last node
        self.nodal_forces[load_id]['N'].append(self.fem_elements[-1].fint[load_id]['fx'][1])
        self.nodal_forces[load_id]['Vz'].append(self.fem_elements[-1].fint[load_id]['fz'][1])
        self.nodal_forces[load_id]['My'].append(self.fem_elements[-1].fint[load_id]['my'][1])
        
        # Determine maximum (design) values of the internal forces
        self.NEd[load_id] = absmax(self.nodal_forces[load_id]['N'])
        self.VzEd[load_id] = absmax(self.nodal_forces[load_id]['Vz'])
        self.MyEd[load_id] = absmax(self.nodal_forces[load_id]['My'])
        
        """
        # i is node index
        i = 0        
        # node_ids is a list of node indices for the member
        node_ids = list(self.nodes.keys())
        node_forces = {}
        for element in self.elements.values():
            node_id = node_ids[i]
            #axial_force = absmax(element.axial_force)
            axial_force = absmax(element.fint[load_id]['fx'])
            if abs(axial_force) > abs(max_ned):
                max_ned = axial_force

            #shear_force = absmax(element.shear_force)
            shear_force = absmax(element.fint[load_id]['fz'])
            if abs(shear_force) > abs(max_ved):
                max_ved = shear_force

            #bending_moment = absmax(element.bending_moment)
            bending_moment = absmax(element.fint[load_id]['my'])
            if abs(bending_moment) > abs(max_med):
                max_med = bending_moment

            
            #self.nodal_forces[node_id] = [axial_force,
                                          shear_force,
                                          bending_moment]
            
            node_forces[node_id] = [axial_force,
                                    shear_force,
                                    bending_moment]
            i += 1

            if i == len(self.elements):

                node_id = node_ids[i]
                #axial_force = absmax(element.axial_force)
                axial_force = absmax(element.fint[load_id]['fx'])
                if abs(axial_force) > abs(max_ned):
                    max_ned = axial_force

                #shear_force = absmax(element.shear_force)
                shear_force = absmax(element.fint[load_id]['fz'])
                if abs(shear_force) > abs(max_ved):
                    max_ved = shear_force

                #bending_moment = absmax(element.bending_moment)
                bending_moment = absmax(element.fint[load_id]['my'])
                if abs(bending_moment) > abs(max_med):
                    max_med = bending_moment

                
                #self.nodal_forces[node_id] = [axial_force,
                                              shear_force,
                                              bending_moment]
                
                node_forces[node_id] = [axial_force,
                                    shear_force,
                                    bending_moment]

        self.nodal_forces[load_id] = node_forces
        self.med = max_med
        self.ved = max_ved
        self.ned = max_ned
        """
    
    def design(self,load_id = LoadIDs['ULS']):
        """ Designs the member (check resistance) """
        
        """ Here, the internal forces of the member at sections must
            somehow be inserted. Note that SteelMember does not yet
            provide several load cases.
        """

        rmax = self.member.design()
        
        if rmax > self.allowable_utilization:
            self.resistance_check = False
        else:
            self.resistance_check = True
            
        self.utilization[load_id] = rmax
    
    def print_utilization(self,file=None,details=False):
        """ Prints utilization ratios for different load cases """
        if file is None:
            print(f' Member {self.mem_id:3.0f}: {self.cross_section.__repr__():s}')
            if details:
                print(f'  Length = {self.length():6.0f} mm')
                print(f'  Buckling Length Factor, y-axis= {self.lcr[0]:4.2f} ')
                print(f'  Buckling Length Factor, z-axis= {self.lcr[1]:4.2f} ')
            for load_id, U in self.utilization.items():
                print(f'  Load case {load_id}: {U:4.3f}')    
        else:
            file.write(f' Member {self.mem_id:3.0f}: {self.cross_section.__repr__():s}\n')
            if details:
                file.write(f'  Length = {self.length():6.0f} mm\n')
                file.write(f'  Buckling Length Factor, y-axis= {self.lcr[0]:4.2f}\n')
                file.write(f'  Buckling Length Factor, z-axis= {self.lcr[1]:4.2f}\n')
            for load_id, U in self.utilization.items():
                file.write(f'  Load case {load_id}: {U:4.3f}\n')
                       
    def optimum_design(self,prof_type,verb):
        """ Searches for the minimum weight profile of 'prof_type' 
            This method must be implemented for each material separately
            because of the profile types.
        """
        
        return False
        
    def opposite_node(self,node):
        """ Returns the node of the member opposite to the 'node'
            in the input argument.
        """
        n = None
        
        for n in self.nodes:
            if n != node:
                break
        
        return n
    
    def generate_elements(self,fem):
        """ Creates internal nodes and elements between nodes 
            Assumes that the member end nodes have already been created
        """
    
        # First end node                    
        self.fem_nodes.append(self.nodes[0].fem_node)
        
        # Generate internal nodes:
        # global_node_coord include also the end node coordinates,
        # so they are not used in the iteration
        for x in self.global_node_coords[1:-1]:  
            if self.frame.dim == 2:
                newNode = fem.add_node(x[0],x[1])
            else:
                newNode = fem.add_node(x[0],x[1],x[2])
                
            self.fem_nodes.append(newNode)            
        
        # Last end node
        self.fem_nodes.append(self.nodes[1].fem_node)
        
        # Create members
        # The zip command allows to create two iterables, where the
        # items are actually consecutive items of the list fem_nodes.
        for n1, n2 in zip(self.fem_nodes,self.fem_nodes[1:]):            
            if self.mtype == 'beam' or self.mtype == 'column':
                #section = framefem.BeamSection(self.cross_section.A,self.cross_section.Iy)
                #material = self.make_framefem_material()
                if self.frame.dim == 2:        
                    newElement = EBBeam(n1,n2,self.cross_section,self.material)
                else:
                    newElement = EBBeam3D(n1,n2,self.cross_section,self.material)
            elif self.mtype == 'bar' or self.mtype == 'brace':
                """ Simple bar element that carries only axial forces """
                #material = self.make_framefem_material()
                newElement = Rod(n1,n2,self.cross_section,self.material)
            elif self.mtype == "top_chord" or self.mtype == "bottom_chord": 
                if self.frame.dim == 2:        
                    newElement = EBBeam(n1,n2,self.cross_section,self.material)
                else:
                    newElement = EBBeam3D(n1,n2,self.cross_section,self.material)
                
            
            fem.add_element(newElement)
            self.fem_elements.append(newElement)
        
        if self.hinges[0]:
            
            """ There is a hinge in the first end """
            self.fem_elements[0].releases = [2]
        
        if self.hinges[1]:
            """ There is a hinge in the last end """
            self.fem_elements[-1].releases = [5]
    
    def clear_fem(self):
        """ Removes FEM nodes and elements """
        self.fem_nodes.clear()
        self.fem_elements.clear()
    
    def update_element_sections(self):
        """ Updates the cross sections of the finite elements of the member """
        if len(self.fem_elements) > 0:
            for element in self.fem_elements:
                element.section = self.cross_section
    
    def plot(self, print_text=True, c='k', axes=None, mem_dim=False, dy = 50):
        
        if self.active:
            if axes is None:
                fig, ax = plt.subplots(1)
            else:
                ax = axes
    
            X0, X1 = self.coords()
            if c:
                if self.resistance_check:
                    color = 'green'
                else:
                    color = 'red'
            else:
                color = 'k'
            # Plot members
            if mem_dim:
                ax.plot([X0[0], X1[0]], [X0[1], X1[1]], color, linestyle='dashdot')
                v = self.perpendicular
                h = 0.5*self.cross_section.H*v
                ax.plot([X0[0]+h[0], X1[0]+h[0]], [X0[1]+h[1], X1[1]+h[1]], color, linestyle='solid')
                ax.plot([X0[0]-h[0], X1[0]-h[0]], [X0[1]-h[1], X1[1]-h[1]], color, linestyle='solid')
            else:
                ax.plot([X0[0], X1[0]], [X0[1], X1[1]], color)
            
            # Plot text
            if self.mtype == 'beam':
                horzalign = 'center'                
                vertalign = 'bottom'
    
            elif self.mtype == 'column':
                horzalign = 'right'
                vertalign = 'center'
    
            else:
                horzalign = 'center'
                vertalign = 'center'
    
            x, y = self.local_to_global(0.5) - 0.5*self.perpendicular * dy
            rot = np.degrees(self.angle)
    
            if rot > 90:
                rot = -(180-rot)
    
            if print_text:                
                ax.text(x, y, str(self.mem_id) + ": " + str(self.cross_section),
                         rotation=rot, rotation_mode='anchor',
                         horizontalalignment=horzalign,
                         verticalalignment=vertalign)
            #else:
            #    ax.text(x, y, str(self.mem_id),
            #             rotation=rot, horizontalalignment=horzalign,
            #             verticalalignment=vertalign)
    
    def plot3d(self, print_text=True, c='k', axes=None, mem_dim=False, dy = 50):
        
        if self.active:
            if axes is None:
                fig, ax = plt.subplots(1,subplot_kw=dict(projection='3d'))
            else:
                ax = axes
    
            X0, X1 = self.coords()
            if c:
                if self.resistance_check:
                    color = 'green'
                else:
                    color = 'red'
            else:
                color = 'k'
            # Plot members
            if mem_dim:
                ax.plot([X0[0], X1[0]], [X0[1], X1[1]], color, linestyle='dashdot')
                v = self.perpendicular
                h = 0.5*self.cross_section.H*v
                ax.plot([X0[0]+h[0], X1[0]+h[0]], [X0[1]+h[1], X1[1]+h[1]], color, linestyle='solid')
                ax.plot([X0[0]-h[0], X1[0]-h[0]], [X0[1]-h[1], X1[1]-h[1]], color, linestyle='solid')
            else:
                # zdir = 'y' indicates that mplot3D package considers the second coordinate
                # as the z coordinate. 
                ax.plot([X0[0], X1[0]], [X0[1], X1[1]], [X0[2], X1[2]], zdir='y', color=color)
            
            """
            # Plot text
            if self.mtype == 'beam':
                horzalign = 'center'
                vertalign = 'bottom'
    
            elif self.mtype == 'column':
                horzalign = 'right'
                vertalign = 'center'
    
            else:
                horzalign = 'center'
                vertalign = 'center'
    
            x, y = self.local_to_global(0.5) - self.perpendicular * dy
            rot = np.degrees(self.angle)
    
            if rot > 90:
                rot = -(180-rot)
    
            if print_text:                
                ax.text(x, y, str(self.mem_id) + ": " + str(self.cross_section),
                         rotation=rot, horizontalalignment=horzalign,
                         verticalalignment=vertalign)
            else:
                ax.text(x, y, str(self.mem_id),
                         rotation=rot, horizontalalignment=horzalign,
                         verticalalignment=vertalign)
            """

    def bmd(self, scale=1, load_id = LoadIDs['ULS'], plot_self=True, show=True):
        """ Plots bending moment diagram """
        if plot_self:
            self.plot(  )
        # Scales Nmm to kNm
        unit_scaler = 1e-6

        # Member's position vector
        #v = self.dir_vector
        # Vector perpendicular to v
        u = -self.perpendicular
        X = []
        Y = []
        start_coord, end_coord = self.coords()
        x01, y01 = start_coord
        x02, y02 = end_coord
        X.append(x01)
        Y.append(y01)
        
        # Get bending moment values for the given load case
        moment_values = self.nodal_forces[load_id]['My']
        
        max_moment = max(moment_values)
        min_moment = min(moment_values)

        for elem in self.fem_elements:
            x0, y0 = elem.nodes[0].coord
            bending_moment = elem.fint[load_id]['my'][0]
            # Scale the bending moment
            val = bending_moment * unit_scaler * scale
            # Add point to the bending moment diagram
            x, y = np.array([x0, y0]) - val * u
            X.append(x)
            Y.append(y)
            horzalign = 'center'
            #vertalign = 'center'
            # Add the value of the maximum and minimum values as text in the figure
            if bending_moment == max_moment or bending_moment == min_moment:
                plt.text(x, y, f'{bending_moment * unit_scaler:.2f} kNm',
                         horizontalalignment=horzalign)
        
        # These points are the last in the list
        x0, y0 = elem.nodes[1].coord
        bending_moment = elem.fint[load_id]['my'][1]
        val = bending_moment * unit_scaler * scale
        x, y = np.array([x0, y0]) - val * u
        X.append(x)
        Y.append(y)
        X.append(x02)
        Y.append(y02)
        plt.text(x, y, f'{bending_moment * unit_scaler:.2f} kNm',
                 horizontalalignment=horzalign)
        plt.plot(X, Y, color='gray')

        if show:
            plt.show()

    def add_hinge(self, loc=0):
        """ Adds a hinge to the member
        input:
            loc .. location along the member (should be 0 or 1)
        """
        if loc == 0:
            self.hinges[0] = True
        else:
            self.hinges[1] = True
    
    def robot_releases(self):
        """ Returns a string indicating releases that Robot Structural Analysis
            can use in creating a str file.
        """
        if self.hinges[0] == True and self.hinges[1] == True:
            s = f'ELEments {self.mem_id+1} ORIgin RY END RY\n'
        elif self.hinges[0] == True and self.hinges[1] == False:
            s = f'ELEments {self.mem_id+1} ORIgin RY \n'
        elif self.hinges[0] == False and self.hinges[1] == True:
            s = f'ELEments {self.mem_id+1} END RY \n'
        else:
            s = None
        return s

class MultiSpanMember:
    """ Class for member with multiple spans """
    
    
    def __init__(self,nodes,section,mem_type="beam",mem_id="",nel=2,hinges=[False,False],lcr=[1.0,1.0],sway=False):
        """
        Constructor

        Parameters
        ----------
        nodes : list of FrameNode objects
            Provides the nodes of the member. Each pair of consecutive nodes constitutes
            one span of the member
        section : SteelSection
            Cross-section of the member.
        mem_type : string, optional
            member type. The default is "beam".
        mem_id : string, optional
            Member id or name. The default is "".
        nel : int, optional
            Number of finite elments per span. The default is 2.
        hinges : TYPE, optional
            DESCRIPTION. The default is [False,False].
        lcr: list, optional
            buckling length factors for the principal axes
            

        Returns
        -------
        None.

        """
        
        self.active = True
        self.resistance_check = True
        self.allowable_utilization = 1.0
        
        self.fem_nodes = []
        self.fem_elements = []
        
        self.symmetry_pair = None # Potential symmetry member pair
        
        self.members = None
        
        # Buckling length factor
        self.lcr = lcr
        
        
        self.nodes = nodes        
        self.cross_section = section
        self.mtype = mem_type
        self.mem_id = mem_id
        self.frame = None
        
        # num_elements is the number of elements per member segment.
        # If 'nel' is given as a scalar value, then a list is made
        # with each element having the value 'nel'.
        if isinstance(nel,list):
            self.num_elements = nel
        else:
            self.num_elements = [nel for i in range(len(nodes)-1)]
            
        self.local_node_coords = [] # Local coordinates of member nodes in FEM model
        self.global_node_coords = [] # Local coordinates of member nodes in FEM model
        
        self.loads = [] # List of loads acting on the member
        
        self.nodal_forces = {} # Nodal forces in different load cases
        self.NEd = {} # Design value of the axial force (maximum) for each load case
        self.VzEd = {} # Design value of the shear force (maximum) for each load case
        self.MyEd = {} # Design value of the bending moment (maximum) for each load case
        self.utilization = {} # Utilization ratio for different load cases
        
        # Create nodal coordinates
        #self.calc_nodal_coordinates(nel)
        
        # Flags for stating whether or not the member ends are hinged.
        
        self.hinges = hinges
        
        self.sway = sway
        
    def __repr__(self):        
        return f"{type(self).__name__} [{self.mem_id}]"
    
    @property
    def cross_section(self):        
        return self.__cross_section
    
    @cross_section.setter
    def cross_section(self,val):
        
        self.__cross_section = val
        if not self.members is None:
            for mem in self.members:
                mem.cross_section = val
        
            self.update_element_sections()
    
    @property
    def lcr(self):
        return self.__lcr
    
    @lcr.setter
    def lcr(self,val):
        self.__lcr = val
        
        if not self.members is None:
            for mem in self.members:
                mem.lcr = val
        
    @property
    def material(self):
        """ Material of the member """
        return self.cross_section.material
    
    def plot(self, print_text=True, c='k', axes=None, mem_dim=False, dy = 50):
        
        if axes is None:
            fig, ax = plt.subplots(1)
        else:
            ax = axes
                            
        for mem in self.members:
            mem.plot(print_text, c, ax, mem_dim, dy)
    
    def bmd(self, scale=1, load_id = LoadIDs['ULS']):
        """ Bending moment diagram """
        
        for mem in self.members:
            mem.bmd(scale,load_id)
        
    
    def coords(self):
        """ Returns list of nodal coordinates """
        x = [np.asarray(node.coords) for node in self.nodes]
        return x
    
    def end_coords(self):
        """ Returns the coordinates of the end nodes """
        x1, x2 = np.asarray([self.nodes[0].coords,self.nodes[-1].coords])
        return x1, x2

    def length(self):
        """ Calculate the length of the member """
        start_node, end_node = self.end_coords()
        return np.linalg.norm(end_node - start_node)
    
    def weight(self):
        """
        Function that calculates member's weight
        :return weight: float, weight of the member
        """        
        return self.cross_section.A * self.length() * self.cross_section.density
    
    @property
    def dir_vector(self):
        """ Direction vector of the member """
        return self.members[0].dir_vector

    @property
    def perpendicular(self):
        """ Unit vector perpendicular to the axis of the member """        
        return self.members[0].perpendicular
    
    @property
    def angle(self):
        """
        Returns member's angle with respect to x-axis in radians
        """
        return self.members[0].angle
    
    def local_to_global(self, loc):
        """
        Returns local coordinate in global coordinates
        :param loc: local coordinate
        :return: global coordinate
        """
        Lx = self.length() * loc
        return self.nodes[0].coords + Lx * self.dir_vector
    
    def global_to_local(self, x):
        """ Returns the local coordinate of the point x along the
            member axis.
            
            It is assumed that 'x' lies on the member axis, this not checked.
        """
        x0 = self.nodes[0].coords
        L = self.length()
        
        if isinstance(x,list):
            xloc = [np.linalg.norm(xi-x0)/L for xi in x]
        else:
            xloc = np.linalg.norm(x-x0)/L
        
        return xloc
            
    
    def calc_nodal_coordinates(self, num_elements=0):
        """ Calculates nodal coordinates along the member 
            
        """
                
        if num_elements > 0:
            self.num_elements = num_elements
        
        self.global_node_coords = []
        
        for mem in self.members:
            mem.calc_nodal_coordinates(num_elements)
            self.global_node_coords += mem.global_node_coords[:-1]
        
        self.global_node_coords.append(self.members[-1].global_node_coords[-1])
        
        self.local_node_coords = self.global_to_local(self.global_node_coords)
    
    def calc_nodal_forces(self,load_id=LoadIDs['ULS']):
        """ Collects the internal forces at the nodes into
            'nodal_forces' attribute which is a dict.
        """
        for mem in self.members:
            mem.calc_nodal_forces(load_id)
        
    def generate_elements(self,fem):
        """ Creates internal nodes and elements between nodes """
    
        for mem in self.members:
            mem.generate_elements(fem)
            self.fem_elements += mem.fem_elements
            
        if self.hinges[0]:
            
            """ There is a hinge in the first end """
            """ NOTE! This is wrong: the for loop above adds the fem elements and nodes
                for the individual members and not to the MultiSpan member
            """
            self.fem_elements[0].releases = [2]
        
        if self.hinges[1]:
            """ There is a hinge in the last end """
            self.fem_elements[-1].releases = [5]    
    
    def clear_fem(self):
        
        for mem in self.members:
            mem.clear_fem()
        
        self.fem_elements.clear()
    
    def update_element_sections(self):
        """ Updates the cross sections of the finite elements of the member """
        for mem in self.members:
            mem.update_element_sections()
    
    
    def robot_releases(self):
        """ Returns a string indicating releases that Robot Structural Analysis
            can use in creating a str file.
        """
        if self.hinges[0] == True and self.hinges[1] == True:
            s = f'ELEments {self.mem_id+1} ORIgin RY END RY\n'
        elif self.hinges[0] == True and self.hinges[1] == False:
            s = f'ELEments {self.mem_id+1} ORIgin RY \n'
        elif self.hinges[0] == False and self.hinges[1] == True:
            s = f'ELEments {self.mem_id+1} END RY \n'
        else:
            s = None
        return s
class SteelFrameMember(FrameMember):
    """ Class for steel frame members """
    
    def __init__(self,nodes,section,mem_type='beam',mem_id="",nel=2,hinges=[False,False],lcr=[1.0,1.0],sway=False):
        """

        Parameters
        ----------
        nodes : TYPE
            DESCRIPTION.
        section : TYPE
            DESCRIPTION.

        Returns
        -------
        None.

        """
        
        
        super().__init__(nodes,section,mem_type,mem_id,nel,hinges,lcr,sway)
                
        self.member = SteelMember(self.cross_section, self.length(), Lcr=lcr, mtype=self.mtype)
    
        self.create_sections()
    
    def __repr__(self):
        
        s = super().__repr__()
        s += ': ' + self.cross_section.__repr__()
        return s
       
    def material_cost(self):
        """ Material cost of the member """
        
        return self.member.cost()
    
    def blasting_cost(self):
        
        return self.frame.workshop.cost_centres['blasting'].cost(self.length())
        
    def painting_cost(self):
    
        return self.frame.workshop.cost_centres['painting'].cost(self.length()*self.cross_section.Au)
    
    def make_framefem_material(self):
        """ Creates a Material class object that can be used to generate material for FrameFem elements """
        
        mat = framefem.Material(E=self.cross_section.E,rho=self.cross_section.density,nu=self.cross_section.material.nu)
        
        return mat
    
    def create_sections(self):
        """ Creates cross sections to be checked """
        
        for xloc in self.local_node_coords:
            self.member.add_section(loc=xloc)
    
    def is_section(self,profile):
        """ Compares the current cross section of the member with
            'profile'.
        """
        
        return (self.cross_section.A-profile.A)**2 + (self.cross_section.Iy-profile.Iy)**2 < 1e-6
        
    
    def design(self,load_id = LoadIDs['ULS']):
        """ Designs the member (check resistance) """
        
        # Assign forces to SteelMember sections
        #self.member.ned = self.nodal_forces[load_id]['N']
        #self.member.vzed = self.nodal_forces[load_id]['Vz']
        #self.member.myed = self.nodal_forces[load_id]['My']

        N = self.nodal_forces[load_id]['N']
        Vz = self.nodal_forces[load_id]['Vz']
        My = self.nodal_forces[load_id]['My']
        
        for loc, n, vz, my in zip(self.local_node_coords, N, Vz, My):
            self.member.sections[loc]['N'] = n
            self.member.sections[loc]['Vz'] = vz
            self.member.sections[loc]['My'] = my

        rmax = self.member.design(sway=self.sway)
        
        if rmax > self.allowable_utilization:
            self.resistance_check = False
        else:
            self.resistance_check = True
            
        self.utilization[load_id] = rmax
    
    def optimum_design(self, prof_type='CURRENT',verb=False,material="S355",
                       sec_class = 3, max_utility=1.0,
                       bmin=10, bmax=1e5, max_slenderness=5.0):
        """ Goes through all profiles in list and stops iterating when 
            cross-section can bear it's loads.
            If member is previously designed, iterating starts from 3 profiles
            before currrent profile.
        """        
        
        # Get list of profiles profiles of chosen type
        profiles = profile_list(self.cross_section,prof_type)
        
        initial_profile = self.cross_section

        for profile in profiles:
            # profile is a string containing the name of the profile
            # It is converted to a SteelSection object using the make_section
            # function of the 'catalogue.py' file.
            self.cross_section = make_section(profile)
            self.material = material
            self.cross_section.Ned = -1
            
            if self.cross_section.b < bmin or self.cross_section.b > bmax:
                continue
            
            if self.cross_section.section_class() > sec_class:
                continue
            
            if any(self.member.slenderness() > max_slenderness):
                continue
            
            for load_id in self.frame.load_ids:
                self.design(load_id)
                
                # Design also symmetry pair, if it exists
                if not self.symmetry_pair is None:
                    self.symmetry_pair.design(load_id)
                
                            
            if all(r<=max_utility for r in self.utilization.values()):# <= 1.0:                            
                # Check also symmetry pair, if it exists
                if not self.symmetry_pair is None:
                    if all(r<=max_utility for r in self.symmetry_pair.utilization.values()):# <= 1.0:                        
                        break
                else:
                    break
        
        """ If the profile changed during iteration, return 'True'. """
        if not self.is_section(initial_profile):
            if verb:
                print(f'Initial profile: {initial_profile}')
                print(f'New profile: {self.cross_section}')
            return True
        else:
            return False


class MultiSpanSteelMember(MultiSpanMember):
    """ Class for steel frame members """
    
    def __init__(self,nodes,section,mem_type='beam',mem_id="",nel=2,hinges=[False,False],lcr=[1.0,1.0],sway=False):
        """

        Parameters
        ----------
        nodes : TYPE
            DESCRIPTION.
        section : TYPE
            DESCRIPTION.

        Returns
        -------
        None.

        """
        super().__init__(nodes,section,mem_type,mem_id,nel,hinges,lcr,sway)
        
        self.members = []
                        
        n = len(self.members)
        for nel, n1, n2 in zip(self.num_elements,nodes[:-1],nodes[1:]):
            new_mem_id = mem_id + str(n)
            n += 1
            self.members.append(SteelFrameMember([n1,n2],section,mem_type,new_mem_id,nel,lcr=lcr,sway=sway))
            self.members[-1].frame = self.frame
    
    def is_section(self,profile):
        """ Compares the current cross section of the member with
            'profile'.
        """
        
        return (self.cross_section.A-profile.A)**2 + (self.cross_section.Iy-profile.Iy)**2 < 1e-6
    
    def print_utilization(self,file=None, details=False):
        """ Prints utilization ratios for different load cases """
        if file is None:
            print(f' Member {self.mem_id:3.0f}: {self.cross_section.__repr__():s}')
            if details:
                print(f'  Length = {self.length():6.0f} mm')
                print(f'  Buckling Length Factor, y-axis= {self.lcr[0]:4.2f} ')
                print(f'  Buckling Length Factor, z-axis= {self.lcr[1]:4.2f} ')
            for load_id, U in self.utilization.items():
                print(f'  Load case {load_id}: {U:4.3f}')    
        else:
            file.write(f' Member {self.mem_id:3.0f}: {self.cross_section.__repr__():s}\n')
            if details:
                file.write(f'  Length = {self.length():6.0f} mm\n')
                file.write(f'  Buckling Length Factor, y-axis= {self.lcr[0]:4.2f}\n')
                file.write(f'  Buckling Length Factor, z-axis= {self.lcr[1]:4.2f}\n')
            for load_id, U in self.utilization.items():
                file.write(f'  Load case {load_id}: {U:4.3f}\n')
    
    def design(self,load_id = LoadIDs['ULS']):
        """ Designs the member (check resistance) """
        
        rmax = 0
        
        for mem in self.members:
            mem.design(load_id)
            rmax = max(rmax,mem.utilization[load_id])
                        
        if rmax > 1.0:
            self.resistance_check = False
        else:
            self.resistance_check = True
            
        self.utilization[load_id] = rmax
    
    def optimum_design(self,prof_type="CURRENT",verb=False,material="S355",
                       sec_class=3,max_utility=1.0):
        """ Goes through all profiles in list and stops iterating when 
            cross-section can bear it's loads.
            If member is previously designed, iterating starts from 3 profiles
            before currrent profile.
        """
        
        #Get list of profiles profiles of chosen type
        profiles = profile_list(self.cross_section,prof_type)        
        initial_profile = self.cross_section        

        for profile in profiles:
            # profile is a string containing the name of the profile
            # It is converted to a SteelSection object using the make_section
            # function of the 'catalogue.py' file.
            self.cross_section = make_section(profile,material)
            
            #print(self.cross_section,self.cross_section.material)
            
            sec = self.cross_section
            sec.Ned = -1
            
            if sec.section_class() > sec_class:
                continue
            
            for load_id in self.frame.load_ids:   
                self.design(load_id)
                            
            if all(r<=self.allowable_utilization for r in self.utilization.values()):# <= 1.0:
                break
            
        
        """ If the profile changed during iteration, return 'True'. """    
        if not self.is_section(initial_profile):
            if verb:
                print(f'Initial profile: {initial_profile}')
                print(f'New profile: {self.cross_section}')
            return True
        else:
            return False

class MemberGroup:
    """ Class for a group of members. Used primarily for setting the cross-section
        and optimization.
    """
    
    def __init__(self,members=[],name="Group"):
        """
        

        Parameters
        ----------
        members : List
            List of initial members.
        name : TYPE, optional
            DESCRIPTION. The default is "Group".

        Returns
        -------
        None.

        """

        self.members = members
        if len(members) == 0:
            self.cross_section = None
            self.frame = None
        else:
            self.cross_section = members[0].cross_section        
            self.frame = members[0].frame
        self.utilization = {}
        self.resistance_check = True
        
        

    @property
    def cross_section(self):        
        return self.__cross_section
    
    @cross_section.setter
    def cross_section(self,val):
        
        self.__cross_section = val
        if len(self.members) > 0:
            for mem in self.members:
                mem.cross_section = val
        
            mem.update_element_sections()
    
    @property
    def material(self):
        """ Material of the member """
        return self.__cross_section.material
    
    @material.setter
    def material(self,val):
        """ Set member material """
        self.__cross_section.material = val
        if len(self.members) > 0:
            for mem in self.members:
                mem.cross_section.material = val
            
    def add_member(self,new_member):
        
        self.members.append(new_member)
        
        if self.cross_section is None:
            self.cross_section = new_member.cross_section
        
        if self.frame is None:
            self.frame = new_member.frame
    
    def is_section(self,profile):
        """ Compares the current cross section of the member with
            'profile'.
        """
                        
        return (self.cross_section.A-profile.A)**2 + (self.cross_section.Iy-profile.Iy)**2 < 1e-6
    
    def design(self,load_id = LoadIDs['ULS']):
        """ Designs the member group (check resistance) """
        
        rmax = 0
        
        for mem in self.members:
            mem.design(load_id)
            rmax = max(rmax,mem.utilization[load_id])
                        
        if rmax > 1.0:
            self.resistance_check = False
        else:
            self.resistance_check = True
            
        self.utilization[load_id] = rmax
    
    def optimum_design(self,prof_type="CURRENT",verb=False,material="S355",
                       sec_class=3,max_utility=1.0,
                       bmin=10, bmax=1e5, max_slenderness=5.0):
        """ Goes through all profiles in list and stops iterating when 
            cross-section can bear it's loads.
            If member is previously designed, iterating starts from 3 profiles
            before currrent profile.
        """
        
        #print(self.cross_section)
        #print(prof_type)
        #Get list of profiles profiles of chosen type        
        profiles = profile_list(self.cross_section,prof_type)
        
        #print(profiles)

        initial_profile = self.cross_section

        for profile in profiles:
            # profile is a string containing the name of the profile
            # It is converted to a SteelSection object using the make_section
            # function of the 'catalogue.py' file.
            self.cross_section = make_section(profile)            
            self.material = material
            sec = self.cross_section
            sec.Ned = -1            
            
            #print(sec)
            
            if self.cross_section.b < bmin or self.cross_section.b > bmax:
                continue
            
            if sec.section_class() > sec_class:
                continue
            
            for member in self.members:
                # For multispan members, the member of a group contains
                # several members.
                if isinstance(member,MultiSpanMember):
                    for mem in member.members:
                        if any(mem.member.slenderness() > max_slenderness):
                            continue
                else:
                    if any(member.member.slenderness() > max_slenderness):
                        continue
            
            
            for load_id in self.frame.load_ids:   
                self.design(load_id)
                            
            if all(r<=max_utility for r in self.utilization.values()):# <= 1.0:
                break
            
        
        """ If the profile changed during iteration, return 'True'. """    
        if not self.is_section(initial_profile):
            if verb:
                print(f'Initial profile: {initial_profile}')
                print(f'New profile: {self.cross_section}')
            return True
        else:
            return False

if __name__ == "__main__":
    
    L = 8000
    
    X = [0, 0.6*L, L]
    Y = [0, 0 ,0]
    
    nodes = []
    for x, y in zip(X,Y):
        nodes.append(FrameNode([x,y]))
    
    section = HEA(260)
    
    m = MultiSpanSteelMember(nodes, section, mem_id="B")
    m.plot(dy=0)
    m.calc_nodal_coordinates()
    
        


# -*- coding: utf-8 -*-
# Copyright 2022 Kristo Mela
# This source code is licensed under the MIT license. See LICENSE in the repository root directory.
# Author(s): Kristo Mela
"""
This is an example for creating and analyzing
a simply supported beam
"""

from metku.frame2d import *

# Create empty frame

frame = Frame2D()
# Create beam
coordinates = [[0, 0], [5000, 0]]
beam = SteelBeam(coordinates,
                 profile="ipe 100",
                 material="S355",
                 num_elements=6)
# Add beam to frame
frame.add(beam)
# Create lineload
lineload = LineLoad(beam, [-10, -10], 'y')
# Add load to frame
frame.add(lineload)
# Create supports
sup1 = XYHingedSupport([0, 0])
sup2 = XYHingedSupport([5000, 0])
# Add supports to frame
frame.add(sup1)
frame.add(sup2)
# Generate FE-model of the frame
frame.generate()
# Run linear analysis for frame
frame.calculate()

## Plotting
# Plot bending moment diagram
frame.bmd(scale=50)
# Plot deflected beam
frame.plot_deflection(scale=1)

## Getting force and displacement values

# Print maximum forces
print(f'Maximum normal force: {beam.ned * 1e-3:.2f} kN')
print(f'Maximum shear force: {beam.ved * 1e-3:.2f} kN')
print(f'Maximum bending moment: {beam.med * 1e-6:.2f} kNm')
# Print forces at each node
for node_id, forces in beam.nodal_forces[LoadIDs.ULS].items():
    N, V, M = forces
    print(f'Forces at node {node_id}: \n'
          f'Normal: {N * 1e-3:.2f} kN '
          f'Shear: {V * 1e-3:.2f} kN '
          f'Moment: {M * 1e-6:.2f} kNm ')

# Print displacements at each node
for node_id, forces in beam.nodal_displacements[LoadIDs.ULS].items():
    dx, dy, r = forces
    print(f'Displacements at node {node_id}: \n'
          f'X: {dx:.2f} mm '
          f'Y: {dy:.2f} mm '
          f'Rotation: {r:.2f} radians ')




# Author(s): Kristo Mela
# This source code is licensed under the MIT license. See LICENSE in the repository root directory.
# Copyright 2022 Kristo Mela
# -*- coding: utf-8 -*-
"""
Created on Fri Nov 27 15:16:22 2020

Calculations from EN 1991-1-4, Wind loads

@author: kmela
"""

import math

# Fundamental value of the basic wind velocity, m/s
vb0 = 21
# Directional factor
cdir = 1.0
# Seasonal factor
cseason = 1.0

# Density of air (kg/m3)
rho_air = 1.25

zmax = 200

def interpolate(X,Y,x0):
    """ Linear interpolation between two points """
    
    return Y[0] + (Y[1]-Y[0])/(X[1]-X[0])*(x0-X[0])

terrain_category = {0: {"z0":0.003, "zmin":1.0},
                    1: {"z0":0.01, "zmin":1.0},
                    2: {"z0":0.05, "zmin":2.0},
                    3: {"z0":0.3, "zmin":5.0},
                    4: {"z0":1.0, "zmin":10.0}}

""" EN 1991-1-4, Table 7.1 """
ext_pressure_walls = {"A": {5: {"cpe10":-1.2, "cpe1":-1.4},
                            1: {"cpe10":-1.2, "cpe1":-1.4},
                            0.25: {"cpe10":-1.2, "cpe1":-1.4}},
                      "B": {5: {"cpe10":-0.8, "cpe1":-1.1},
                            1: {"cpe10":-0.8, "cpe1":-1.1},
                            0.25: {"cpe10":-0.8, "cpe1":-1.1}},
                      "C": {5: {"cpe10":-0.5, "cpe1":-0.5},
                            1: {"cpe10":-0.5, "cpe1":-0.5},
                            0.25: {"cpe10":-0.5, "cpe1":-0.5}},
                      "D": {5: {"cpe10":+0.8, "cpe1":1.0},
                            1: {"cpe10":+0.8, "cpe1":1.0},
                            0.25: {"cpe10":+0.7, "cpe1":1.0}},
                      "E": {5: {"cpe10":-0.7, "cpe1":-0.7},
                            1: {"cpe10":-0.5, "cpe1":-0.5},
                            0.25: {"cpe10":-0.3, "cpe1":-0.3}},
                      }

""" EN 1991-1-4, Table 7.2 """
ext_pressure_roof_sharp_eaves = {"F": {"cpe10":-1.8, "cpe1":-2.5},
                            "G": {"cpe10":-1.2, "cpe1":-2.0},
                            "H": {"cpe10":-0.7, "cpe1":-1.2},
                            "I": {"cpe10":[0.2,-0.2], "cpe1":[0.2,-0.2]}}

""" EN 1991-1-4, Table 7.3a """
# TODO!
ext_pressure_monopitch_roof = {0: {5: {"F": {"cpe10":-1.8, "cpe1":-2.5},
                                       "G": {"cpe10":-1.2, "cpe1":-2.0},
                                       "H": {"cpe10":-0.7, "cpe1":-1.2},
                                       "I": {"cpe10":[0.2,-0.2], "cpe1":[0.2,-0.2]}} }}


class ExternalPressureMonoRoof:
    
    def __init__(self,wind_direction=0,angle=5.0):
        
        if angle >= 75:
            if wind_direction == 0:
                self.F = {'cpe10': 0.8, 'cpe1':0.8} 
                self.G = {'cpe10': 0.8, 'cpe1':0.8} 
                self.H = {'cpe10': 0.8, 'cpe1':0.8} 
            elif wind_direction == 180:
                self.F = {'cpe10': -0.5, 'cpe1':-1.0} 
                self.G = {'cpe10': -0.5, 'cpe1':-0.5}
                self.H = {'cpe10': -0.5, 'cpe1':-0.5}
        elif angle >= 60:
            if wind_direction == 0:
                c = interpolate([60,75],[0.7,0.8],angle)
                self.F = {'cpe10': c, 'cpe1':c} 
                self.G = {'cpe10': c, 'cpe1':c} 
                self.H = {'cpe10': c, 'cpe1':c} 
            elif wind_direction == 180:
                self.F = {'cpe10': -0.5, 'cpe1':-1.0} 
                self.G = {'cpe10': -0.5, 'cpe1':-0.5}
                self.H = {'cpe10': -0.5, 'cpe1':-0.5}
        elif angle >= 45:
            if wind_direction == 0:
                c = interpolate([45,60],[0.6,0.7],angle)
                self.F = {'cpe10': [0,0.7], 'cpe1':[0,0.7]} 
                self.G = {'cpe10': [0,0.7], 'cpe1':[0,0.7]} 
                self.H = {'cpe10': [0,c], 'cpe1':[0,c]} 
            elif wind_direction == 180:
                self.F = {'cpe10': interpolate([45,60],[-0.6,-0.5],angle), 'cpe1':interpolate([45,60],[-1.3,-1.0],angle)} 
                self.G = {'cpe10': -0.5, 'cpe1':-0.5}
                c = interpolate([45,60],[-0.7,-0.5],angle)
                self.H = {'cpe10': c, 'cpe1':c}
        elif angle >= 30:
            if wind_direction == 0:
                cneg10 = interpolate([30,45],[-0.5,0.0],angle)
                cneg1 = interpolate([30,45],[-1.5,0.0],angle)
                self.F = {'cpe10': [cneg10,0.7], 'cpe1':[cneg1,0.7]} 
                self.G = {'cpe10': [cneg10,0.7], 'cpe1':[cneg1,0.7]} 
                cpos = interpolate([30,45],[0.4,0.6],angle)
                cneg = interpolate([30,45],[-0.2,0.0],angle)
                self.H = {'cpe10': [cneg,cpos], 'cpe1':[cneg,cpos]} 
            elif wind_direction == 180:
                self.F = {'cpe10': interpolate([30,45],[-1.1,-0.6],angle), 'cpe1':interpolate([30,45],[-2.3,-1.3],angle)} 
                self.G = {'cpe10': interpolate([30,45],[-0.8,-0.5],angle), 'cpe1':interpolate([30,45],[-1.5,-0.5],angle)}
                c = interpolate([30,45],[-0.8,-0.7],angle)
                self.H = {'cpe10': c, 'cpe1':c}
        elif angle >= 15:
            if wind_direction == 0:
                cneg10 = interpolate([15,30],[-0.9,-0.5],angle)
                cneg1 = interpolate([15,30],[-2.0,-1.5],angle)
                cpos = interpolate([15,30],[0.2,0.7],angle)
                self.F = {'cpe10': [cneg10,cpos], 'cpe1':[cneg1,cpos]} 
                cneg10 = interpolate([15,30],[-0.8,-0.5],angle)
                self.G = {'cpe10': [cneg10,cpos], 'cpe1':[-1.5,cpos]} 
                cpos = interpolate([15,30],[0.2,0.4],angle)
                cneg = interpolate([15,30],[-0.3,-0.2],angle)
                self.H = {'cpe10': [cneg,cpos], 'cpe1':[cneg,cpos]} 
            elif wind_direction == 180:
                self.F = {'cpe10': interpolate([15,30],[-2.5,-1.1],angle), 'cpe1':interpolate([15,30],[-2.8,-2.3],angle)} 
                self.G = {'cpe10': interpolate([15,30],[-1.3,-0.8],angle), 'cpe1':interpolate([15,30],[-2.0,-1.5],angle)}
                self.H = {'cpe10': interpolate([15,30],[-0.9,-0.8],angle), 'cpe1':interpolate([15,30],[-1.2,-0.8],angle)}
        elif angle >= 5:
            if wind_direction == 0:
                cneg10 = interpolate([15,30],[-0.9,-0.5],angle)
                cneg1 = interpolate([15,30],[-2.0,-1.5],angle)
                cpos = interpolate([15,30],[0.2,0.7],angle)
                self.F = {'cpe10': [cneg10,cpos], 'cpe1':[cneg1,cpos]} 
                cneg10 = interpolate([15,30],[-0.8,-0.5],angle)
                self.G = {'cpe10': [cneg10,cpos], 'cpe1':[-1.5,cpos]} 
                cpos = interpolate([15,30],[0.2,0.4],angle)
                cneg = interpolate([15,30],[-0.3,-0.2],angle)
                self.H = {'cpe10': [cneg,cpos], 'cpe1':[cneg,cpos]} 
            elif wind_direction == 180:
                self.F = {'cpe10': interpolate([5,15],[-2.3,-2.5],angle), 'cpe1':interpolate([5,15],[-2.5,-2.8],angle)} 
                self.G = {'cpe10': -1.3, 'cpe1':-2.0}
                self.H = {'cpe10': interpolate([5,15],[-0.8,-0.9],angle), 'cpe1':-1.2}
                
                
            self.Fup = 0.0
            self.Flow = 0.0
            self.I = 0.0


def basic_wind_velocity(vb0=21.0,cdir=1.0,cseason=1.0):
    """ EN 1991-1-4, Eq. (4.1) """
    return vb0*cdir*cseason

def terrain_roughness(z,cat=2):
    """ EN 1991-1-4, Eq. (4.4) 
    
        :param z: height (m)
        :param cat: terrain category (0,..,4)
    """    
    z02 = 0.05
    z0  = terrain_category[cat]["z0"]
    zmin = terrain_category[cat]["zmin"]
    
    if cat == 0:
        """ Finnish Natioanl Annex """
        kr = 0.18
    else:
        kr = 0.19*pow(z0/z02,0.07)
    
    return kr*math.log(max(zmin,z)/z0)
    
    
def mean_wind_velocity(vb,z,cat=2):
    """ EN 1991-1-4, Eq. (4.3) """
    cr = terrain_roughness(z,cat)
    # Orography factor
    co = 1.0
    
    return vb*cr*co

def basic_velocity_pressure(vb,rho=1.25):
    """ EN 1991-1-4, Eq. (4.10) """
    return 0.5*rho*vb**2

def turbulence_intensity(z,cat=2):
    """ EN 1991-1-4, Eq. (4.7) """
    # Turbulence factor 
    kI = 1.0
    # Orography factor
    co = 1.0
    
    z0  = terrain_category[cat]["z0"]
    zmin = terrain_category[cat]["zmin"]
    
    return kI/co/(math.log(max(zmin,z)/z0))

def peak_velocity_pressure(z,cat,vb):
    """ EN 1991-1-4, Eq. (4.8) 
        :param z: height (m)
        :param cat: terrain category
        :param vb: basic wind velocity
        
        Output: peak velocity pressure (N/m2)
    """
    
    Iv = turbulence_intensity(z,cat)
    vm = mean_wind_velocity(vb,z,cat)
    
    return (1 + 7*Iv)*0.5*rho_air*vm**2
    
def wind_pressure_ext(qp,cpe):
    """ EN 1991-1-4, Eq. (5.1) """
    
    return qp*cpe

def wind_pressure_int(qp,cpi):
    """ EN 1991-1-4, Eq. (5.2) """
    
    return qp*cpi

def rect_building_wall_pressure(length,width,h,qp,wind_dir="long"):
    """ Calculates external wind pressures in various zones
        for a rectangular building.
        
        EN 1991-1-4, Figure 7.5
    
        :param length: length of the building
        :param width: width of the building
        :param h: height of the building
        :param qp: peak velocity pressure
        :param wind_dir: wind direction.
                        "long" .. wind to the long side
                        "short" .. wind to the short side
    """
    
    if wind_dir == "long":
        b = length
        d = width
    elif wind_dir == "short":
        b = width
        d = length
    
    e = min(b,2*h)
    
    """ Define a dict for the wind pressures in the various
        zones of the wall
    """
    zones = {"A":{"length":0.0,'cpe':0.0,'we':0.0},
             "B":{"length":0.0,'cpe':0.0,'we':0.0},
             "C":{"length":0.0,'cpe':0.0,'we':0.0},
             "D":{"length":b,'cpe':0.0,'we':0.0},
             "E":{"length":b,'cpe':0.0,'we':0.0},
             }
    
    if e < d:
        zones["A"]['length'] = e/5
        zones["B"]['length'] = 0.8*e
        zones["C"]['length'] = d-e
    elif e >= 5*d:
        zones["A"]['length'] = d
    else:
        """ e >= d, but e < 5*d """
        zones["A"]['length'] = 0.2*e
        zones["B"]['length'] = d-0.2*e
    
    """ Area of the wind """
    A = b*h
    
    hd_ratio = h/d
    
    print("*** Wind load ***")
    print("Wind to " + wind_dir + " side of the building")
    print("b = {0:4.2f} m".format(b))
    print("d = {0:4.2f} m".format(d))
    print("h = {0:4.2f} m".format(h))
    print("e = {0:4.2f} m".format(e))
    print("h/d = {0:4.2f} ".format(hd_ratio))
    
    if A > 10:
        zones["A"]['cpe'] = ext_pressure_walls["A"][5]["cpe10"]
        zones["B"]['cpe'] = ext_pressure_walls["B"][5]["cpe10"]
        zones["C"]['cpe'] = ext_pressure_walls["C"][5]["cpe10"]

        if hd_ratio >= 5:
            zones["D"]['cpe'] = ext_pressure_walls["D"][5]["cpe10"]
            zones["E"]['cpe'] = ext_pressure_walls["E"][5]["cpe10"]
        elif hd_ratio >= 1:
            """ Linear interpolation for E """
            zones["D"]['cpe'] = ext_pressure_walls["D"][1]["cpe10"]
            Y = [ext_pressure_walls["E"][1]["cpe10"],
                 ext_pressure_walls["E"][5]["cpe10"]]
            X = [1,5]            
            zones["E"]['cpe'] = interpolate(X,Y,hd_ratio)
        elif hd_ratio >= 0.25:
            """ Linear interpolation for D and E """            
            X = [0.25,1]                        
            YD = [ext_pressure_walls["D"][0.25]["cpe10"],
                 ext_pressure_walls["D"][1]["cpe10"]]            
            YE = [ext_pressure_walls["E"][0.25]["cpe10"],
                 ext_pressure_walls["E"][1]["cpe10"]]            
            zones["D"]['cpe'] = interpolate(X,YD,hd_ratio)
            zones["E"]['cpe'] = interpolate(X,YE,hd_ratio)
            
            #zones["E"]['cpe'] = ext_pressure_walls["E"][1]["cpe10"]
        else:
            zones["D"]['cpe'] = ext_pressure_walls["D"][0.25]["cpe10"]
            zones["E"]['cpe'] = ext_pressure_walls["E"][0.25]["cpe10"]

    for key, value in zones.items():
        value['we'] = qp*value['cpe']
        
    return zones

def monopitch_roof_pressure(length,width,h,qp,wind_angle=180,alpha=5):
    """ Calculates external wind pressures in various zones
        for a rectangular building.
        
        EN 1991-1-4, Figure 7.5
    
        :param length: length of the building
        :param width: width of the building
        :param h: height of the building
        :param qp: peak velocity pressure
        :param wind_anlge: wind direction.
                        0 .. up the slope
                        180 .. against the slope
                        90 .. perpendicular to the slope
    """
    
    if wind_angle == 0 or wind_angle == 180:
        b = length
        d = width
    else:
        b = width
        d = length
    
    if wind_angle == 0:
        wind_dir = 'low eave'
    elif wind_angle == 180:
        wind_dir = 'high eave'
    else:
        wind_dir = 'short'
    
    e = min(b,2*h)
    
    """ Define a dict for the wind pressures in the various
        zones of the wall
    """
    zones = {"F":{"length":0.25*e,'width':0.1*e,'cpe':0.0,'we':0.0},
             "G":{"length":0.5*e,'width':0.1*e,'cpe':0.0,'we':0.0},
             "H":{"length":b,'width':0.9*e,'cpe':0.0,'we':0.0},
             "I":{"length":b,'width':0.0,'cpe':0.0,'we':0.0},
             "Fup":{"length":0.25*e,'width':0.1*e,'cpe':0.0,'we':0.0},
             "Flow":{"length":0.25*e,'width':0.1*e,'cpe':0.0,'we':0.0},
             }
    
    if wind_angle == 90:
        zones['H']['width'] = 0.4*e
        zones['I']['width'] = length-0.5*e
            
    
    """ Area of the wind """
    A = b*h
    
    hd_ratio = h/d
    
    print("*** Wind load ***")
    print("Wind to " + wind_dir + " side of the building")
    print("b = {0:4.2f} m".format(b))
    print("d = {0:4.2f} m".format(d))
    print("h = {0:4.2f} m".format(h))
    print("e = {0:4.2f} m".format(e))
    
    
    ep = ExternalPressureMonoRoof(wind_angle,alpha)
    
   

    for key, value in zones.items():
        value['we'] = qp*value['cpe']
        
    return zones

def flat_roof_pressure(d,b,h,qp,hp=0):
    """ Calculates external wind pressures in various zones
        for a flat roof.
        
        EN 1991-1-4, Figure 7.6
    
        :param d: length of the building in wind direction
        :param b: length of the building in direction perpendicular to wind
        :param h: height of the building (to eaves)
        :param qp: peak velocity pressure
        :param hp: length of parapets (hp = 0 means no parapets)
    """
    
    e = min(b,2*h)
    
    """ Define a dict for the wind pressures in the various
        zones of the wall
    """
    zones = {"F":{"length":0.25*e,'width':0.1*e,'cpe':0.0,'we':0.0},
             "G":{"length":b-0.5*e,'width':0.1*e,'cpe':0.0,'we':0.0},
             "H":{"length":b,'width':0.4*e,'cpe':0.0,'we':0.0},
             "I":{"length":b,'width':d-0.5*e,'cpe':0.0,'we':0.0}
             }
    
    hd_ratio = h/d
    
    print("*** Wind load ***")
    print("Wind to flat roof")
    print("qp = {0:4.2f} ".format(qp))
    print("b = {0:4.2f} m".format(b))
    print("d = {0:4.2f} m".format(d))
    print("h = {0:4.2f} m".format(h))
    print("e = {0:4.2f} m".format(e))
    print("h/d = {0:4.2f} m".format(hd_ratio))
    
    if hp == 0:
        for key, zone in zones.items():
            zones[key]['cpe'] = ext_pressure_roof_sharp_eaves[key]["cpe10"]
            if isinstance(zone['cpe'],list):
                zones[key]['we'] = [qp*cpe for cpe in zone['cpe']]
            else:
                zones[key]['we'] = qp*zone['cpe']
        
    return zones

#if __name__ == "__main__":

    #wind = rect_building_wall_pressure(length=72,width=24,h=8,qp=qp,wind_dir="long")
    #wind = rect_building_wall_pressure(length=15,width=12,h=7,qp=qp,wind_dir="long")
    #wind = rect_building_wall_pressure(length=30,width=18,h=H,qp=qp,wind_dir="long")
    #wind = rect_building_wall_pressure(length=12*6,width=5*4.8,h=H,qp=qp,wind_dir="short")

    #print(qp)    


# -*- coding: utf-8 -*-
# Copyright 2022 Kristo Mela
# This source code is licensed under the MIT license. See LICENSE in the repository root directory.
# Author(s): Kristo Mela
from tables_and_tuples import *
from cost_data import *

# Cost calculations according to thesis:
# Feature-Based Costing Method for Skeletal Steel
# Structures based on the Process Approach, Jaakko Haapio, 2012

# Additional cost calculation functions according to Diaz et al.
# Optimum design of semi-rigid connections using metamodels, 2012


# Cost of material (steel parts, bolts, nuts and washers)
class MaterialCost:
    NOT_IMPLEMENTED = "NOT IMPLEMENTED YET! Add this feature: "

    def __init__(self):
        self.plate_unit_cost = 0.0
        self.steel_grade_add_on = 0.0
        self.thickness_add_on = 0.0
        self.quantity_add_on = 0.0
        self.quantity_per_thickness_add_on = 1.0
        self.inspection_add_on = 1.0

        self.plate = 0.0                                # Material cost of plate
        self.bolts = 0.0                                # Material cost of bolts, nuts and washers'

        self.total = 0.0

    # Chapter 4.5.2 [Haapio]
    def cost_of_plate(self, cn, sd):

        b_p = sd.plate_width                    # [mm]
        h_p = sd.plate_height                   # [mm]
        t_p = sd.plate_thickness                # [mm]
        volume = b_p*h_p*t_p                    # [mm^3]
        material = sd.plate_material            # [S***]
        cost_function = cn.cost_function

        if cost_function == 'Diaz':
            k_s = Diaz['k_s']                   # [e/kg]
            rho = Diaz['rho']*1.0e-9            # [kg/mm^3]
            self.plate = k_s*volume*rho
        else:
            rho = mat[material]["rho"]                                      # Material density [kg/m^3]
            self.plate_unit_cost = c_SMBPl*(1.0/1000.0)                     # [e/kg]
            self.steel_grade_add_on = c_SMG(material)*(1.0/1000.0)          # [e/kg]
            self.thickness_add_on = c_SMT(t_p)*(1.0/1000.0)                 # [e/kg]

            # For whole structure, used add-on 0.0 for connection cost calculations
            self.quantity_add_on = 0.0                                      # [e/kg]
            self.quantity_per_thickness_add_on = 0.0                        # [e/kg]
            self.inspection_add_on = 0.0                                    # [e/kg]

            # Material cost of plate [e/kg]
            self.plate = volume*rho*(self.plate_unit_cost + self.steel_grade_add_on + self.thickness_add_on +
                                     self.quantity_add_on + self.quantity_per_thickness_add_on + self.inspection_add_on)


def bolt_cost(row, cost_function='Haapio'):

    n_bolts = row.n_row             # Number of bolts in row [pcs]

    d = 0
    if row.d in bolt_sizes:
        d = row.d                       # Nominal bolt size [mm]
    else:
        for i in range(0, len(bolt_sizes)):
            if bolt_sizes[i] > row.d and i != 1:
                d = bolt_sizes[i]
            elif bolt_sizes[i] > row.d and i > 1:
                d = bolt_sizes[i - 1]
            else:
                d = bolt_sizes[-1]
    L = row.L                       # Bolt length [mm]

    if cost_function == 'Diaz':
        k_b_mt = 3.076*(d*0.1)**2.0 - 7.373*(d*0.1) + 4.62      # Bolt unit cost + nut and two washers
        row.material_cost = n_bolts*k_b_mt
    else:
        if row.thread_in_shear_plane:
            row.bolt_unit_cost = bolt_unit_cost_full_thread[L][d]               # [e/pcs]
        else:
            row.bolt_unit_cost = bolt_unit_cost_part_thread[L][d]               # [e/pcs]
        if row.bolt_unit_cost == "N/A":
            row.bolt_unit_cost = bolt_unit_cost_part_thread['max']
            print("Size-length (M{0}-{1}) combination not in cost data tables, max bolt unit cost {2} used!"
                  .format(d, L, row.bolt_unit_cost))
        row.nut_unit_cost = nut_unit_cost[d]                                    # [e/pcs]
        if row.d_washer != 0.0:
            row.washer_unit_cost = washer_unit_cost[d]                          # [e/pcs]
        else:
            row.washer_unit_cost = 0.0

        # Total cost of bolts, nuts and washers
        row.material_cost = n_bolts*(row.bolt_unit_cost + row.nut_unit_cost + row.washer_unit_cost)


# Annual cost, Haapio, eq (2)
# Uniform series end-of-period installment [e/a]
def annual_cost(P, I, n):
    # P = a sum of money invested in the initial year less resale value [e]
    # I = interest rate (cost of capital) [%]
    # n = time, the number of units (years) over which interest accumulates

    return P*((I*(1.0 + I)**n)/((1.0 + I)**n - 1.0))       # [e/a]


# General form for cost of cost center
class CostCenter:
    NOT_IMPLEMENTED = "NOT IMPLEMENTED YET! Add this feature: "

    def __init__(self, T_Nk=0.0, T_Pk=0.0, c_Lk=0.0, c_Eqk=0.0, c_Mk=0.0, c_REk=0.0, c_Sek=0.0, c_Ck=0.0,
                 c_Enk=0.0, C_Ck=0.0, u_k=1.0):
        self.T_Nk = T_Nk
        self.T_Pk = T_Pk
        self.c_Lk = c_Lk
        self.c_Eqk = c_Eqk
        self.c_Mk = c_Mk
        self.c_REk = c_REk
        self.c_Sek = c_Sek
        self.c_Ck = c_Ck
        self.c_Enk = c_Enk
        self.C_Ck = C_Ck
        self.u_k = u_k

        # Total cost of cost center [e]
        self.cost = (T_Nk + T_Pk)*(c_Lk + c_Eqk + c_Mk + c_REk + c_Sek)/u_k + T_Pk*(c_Ck + c_Enk) + C_Ck

    def info(self):
        print(self.T_Nk)
        print(self.T_Pk)
        print(self.c_Lk)
        print(self.c_Eqk)
        print(self.c_Mk)
        print(self.c_REk)
        print(self.c_Sek)
        print(self.c_Ck)
        print(self.c_Enk)
        print(self.C_Ck)
        print(self.u_k)


# Cost center for blasting
class BlastingCost(CostCenter):

    def __init__(self, cn, sd):

        cost_function = cn.cost_function
        L_b = max(sd.plate_width, sd.plate_height)      # [mm] (max dimension of part to be blasted)

        if cost_function == 'Diaz':
            self.cost = 0.0
        else:
            L = 40.0                                # Blasting workspace length [m]
            B = 10.0                                # Blasting workspace width [m]
            Area = L*B                              # Workspace area [m^2]

            # Non-productive time
            T_Nk = 0.0

            # Productive time
            v_c = 3000.0                            # Conveyor speer [mm/m]
            T_Pk = L_b/v_c                          # Total processing time [min]

            # Labour
            c_hour = 27.68                          # Hourly cost [e/h]
            n_worker = 1.0                          # Number of workers [-]
            c_min = c_hour/60.0                     # Minute rate of worker [e/min]
            c_Lk = n_worker*c_min                   # Unit labor cost [e/min]

            # Equipment installment cost [e/min]
            P_invest_Eq = 200000.0                  # Investment of equipment [e]
            P_resale_Eq = 0.0                       # Resale value [e]
            I_Eq = 0.05                             # Interest rate (cost of capital) [%]
            n_Eq = 20.0                             # n = time [years]
            c_Eqk = annual_cost(P_invest_Eq - P_resale_Eq, I_Eq, n_Eq)/work_mins_in_a

            # Maintenance
            c_maintenance_Eq = 1000.0                   # Maintenance cost of equipment [e/a]
            c_Mk = c_maintenance_Eq/work_mins_in_a      # Maintenance cost of equipment [e/min]

            # Real estate installment cost [e/min]
            P_invest_RealS = Area*c_construction    # Investment cost of workspace area [e]
            P_resale_RealS = Area*c_land            # Resale value of workspace [e]
            I_RealS = 0.05                          # Interest rate (cost of capital) [%]
            n_RealS = 50.0                          # n = time (years)
            c_REk = annual_cost(P_invest_RealS - P_resale_RealS, I_RealS, n_RealS)/work_mins_in_a

            # Real estate maintenance cost
            c_maintenance_RealS = Area*RE_maintenance             # Real estate maintenance cost[e]
            c_Sek = c_maintenance_RealS/work_mins_in_a            # Real estate maintenance cost[e/min]

            # Consumables: grains
            grain_consump = 0.042                   # Grain consumption on blasting [kg/min], equipment utilization ratio 1
            grain_unit_cost = 0.50                  # Grain weight price [e/kg]
            c_Ck = grain_consump*grain_unit_cost    # Cost of consumables [e/min]

            # Energy
            P_total = 40                            # Total power consumption using four blasting turbines [kW]
            c_En_unit = 0.1                         # Energy unit cost [e/kWh]
            c_Enk = P_total*(c_En_unit/60.0)        # Cost of energy [e/min]

            C_Ck = 0.0                              # Total cost of non-time-related consumables

            u_k = 1.0                               # Workspace utilization ratio, set to 1

            CostCenter.__init__(self, T_Nk, T_Pk, c_Lk, c_Eqk, c_Mk, c_REk, c_Sek, c_Ck, c_Enk, C_Ck, u_k)


# Cost center for cutting
class CuttingCost(CostCenter):

    def __init__(self, cn, sd, n_holes=0, L_holes=0.0):

        t_p = sd.plate_thickness                            # [mm]
        t_f = cn.column.t_f                                 # [mm]
        d_0 = sd.d_0                                        # [mm]
        n_holes = n_holes                                   # [-]
        L_holes = L_holes                                   # [mm]
        L_plate = 2.0*(sd.plate_width + sd.plate_height)    # [mm]
        cost_function = cn.cost_function

        if cost_function == 'Diaz':
            # End-plate thickness, flange thickness and bolt nominal size in cm in calculations

            # Cost of cutting the end-plate to size
            k_c = Diaz['k_c']                               # Unit cost for cutting [e/min]
            f_c = Diaz['f_c']                               # Cutting factor that increases labor [-]
            L_c = L_plate*0.001                             # Cutting length [m]
            T_c_ex = Diaz['T_c_ex']                                 # Additional cutting time
            T_c = -0.015*(t_p*0.1)**2.0 + 0.421*(t_p*0.1) + 1.43    # Cutting time [min/m]
            plate_cutting = k_c*(f_c*T_c*L_c + T_c_ex)              # Total cost of cutting [e]

            # Cost of oxygen propane mixture used for cutting the end-plate
            M_c_o = 1.645*(t_p*0.1)**2.0 + 56.644*(t_p*0.1) - 6.73      # Material consumption of oxygen [l/m]
            M_c_p = 2.171*(t_p*0.1) + 7.87                              # Material consumption of propane [l/m]
            k_c_m_o = Diaz['k_c_m_o']                           # Unit cost of oxygen [e/l]
            k_c_m_p = Diaz['k_c_m_p']                           # Unit cost of propane [e/l]
            gas_cost = L_c*(k_c_m_o*M_c_o + k_c_m_p*M_c_p)      # Total cost of cutting gas [e]

            # Hole forming cost (drilling)
            V_d = 0.763*(d_0*0.1)**2.0 - 5.720*(d_0*0.1) + 20.96    # Drillig speed in progression [cm/min]
            t = (t_p + t_f)*0.1                                     # Total drilling thickness [cm]
            k_h = Diaz['k_h']                                       # Unit cost of hole forming [e]
            n_h = n_holes                                           # Number of holes [-]
            t_ex = Diaz['t_ex']                                     # Additional drilling path [cm]
            T_h_ex = Diaz['T_h_ex']                                 # Additional hole forming time [min]
            hole_forming = k_h*(n_h*(t + t_ex)/V_d + T_h_ex)        # Total cost of hole forming [e]

            self.cost = plate_cutting + gas_cost + hole_forming     # Total cost of end-plate shaping
        else:
            L = 20.0            # Workspace length [m]
            B = 10.0            # Workspace width [m]
            Area = L*B          # Workspace area [m^2]

            # Hole cutting on column flange taken into account with multiplier 2
            L_cut = 2.0*L_holes + L_plate

            # Non-productive time [min]
            T_Nk = 3.0

            # Productive time [min]
            # Cutting speed according to thickness [mm] of the plate
            if t_p <= 30.0:   # Plasma cutting
                v_cut = (8.9212*t_p**2.0 - 486.87*t_p + 8155.8)     # [mm/min]
            else:   # Flame cutting
                v_cut = -4.1939*t_p + 658.67                        # [mm/min]
            T_Pk = L_cut/v_cut                                      # [min]

            # Labour
            c_hour = 27.68                  # Hourly cost [e/h]
            n_worker = 2.0                  # Number of workers [-]
            c_min = c_hour / 60.0           # Minute rate of worker [e/min]
            c_Lk = n_worker * c_min         # Unit labor cost [e/min]

            # Equipment installment cost [e/min]
            P_invest_Eq = 280000.0          # Investment of equipment [e]
            P_resale_Eq = 0.0               # Resale value [e]
            I_Eq = 0.05                     # Interest rate (cost of capital) [%]
            n_Eq = 20.0                     # n = time [years]
            c_Eqk = annual_cost(P_invest_Eq - P_resale_Eq, I_Eq, n_Eq)/work_mins_in_a

            # Maintenance
            c_maintenance_Eq = 1000.0                       # Maintenance cost of equipment [e/a]
            c_Mk = c_maintenance_Eq/work_mins_in_a        # Maintenance cost of equipment [e/min]

            # Real estate installment cost [e/min]
            P_invest_RealS = Area * c_construction          # Investment cost of workspace area [e]
            P_resale_RealS = Area * c_land                  # Resale value of workspace [e]
            I_RealS = 0.05                                  # Interest rate (cost of capital) [%]
            n_RealS = 50.0                                  # n = time (years)
            c_REk = annual_cost(P_invest_RealS - P_resale_RealS, I_RealS, n_RealS)/work_mins_in_a

            # Real estate maintenance cost
            c_maintenance_RealS = Area * RE_maintenance         # Real estate maintenance cost[e]
            c_Sek = c_maintenance_RealS/work_mins_in_a        # Real estate maintenance cost[e/min]

            # Consumables: Gases, plasma electrodes and nozzles
            if t_p <= 30.0:   # Plasma cutting
                # Total consumables cost [e/min]
                cutting_oxygen = 0.071                                       # [m^3/min]
                oxygen_cost = 0.3                                            # [e/min]
                nozzle_unit_cost = 40                                        # [e]
                nozzle_life = 480                                            # [min]
                nozzle_cost = nozzle_unit_cost/nozzle_life                   # [e/min]
                c_Ck = cutting_oxygen*oxygen_cost + nozzle_cost

                # Plasma cutting, Total power consumption [kW]
                P_total = 72.0
            else:   # Flame cutting
                # Total consumables cost [e/min]
                propane_consump = 0.0063                                        # [m^3/min]
                propane_cost = 18.40                                            # [e/m^3]
                preheat_oxygen_consump = 0.025                                  # [m^3/min]
                cutting_oxygen_consump = 1.0e-5*t_p**2.0 + 0.001*t_p + 0.0224   # [m^3/min]
                oxygen_cost = 4.18                                              # [e/m^3]
                c_Ck = propane_consump*propane_cost + (preheat_oxygen_consump + cutting_oxygen_consump)*oxygen_cost

                # Flame cutting, Total power consumption [kW]
                P_total = 0.0

            c_Enk = P_total*(c_En_unit/60.0)        # Cost of energy [e/min]

            C_Ck = 0.0      # Total cost of non-time-related consumables

            u_k = 1.0       # Workspace utilization ratio, set to 1

            CostCenter.__init__(self, T_Nk, T_Pk, c_Lk, c_Eqk, c_Mk, c_REk, c_Sek, c_Ck, c_Enk, C_Ck, u_k)


class DrillingCost(CostCenter):
    NOT_IMPLEMENTED = "NOT IMPLEMENTED YET! Add this feature: "

    def __init__(self, cn, sd):

        cost_function = cn.cost_function

        if cost_function == 'Diaz':
            self.cost = 0.0
        else:
            L = 35.0            # Workspace length [m]
            B = 15.0            # Workspace width [m]
            Area = L * B        # Workspace area [m^2]

            # Non-productive time of cost center [min], setup and moving
            T_Nk = 0.0
            T_Pk = 0.0
            c_Lk = 0.0
            c_Eqk = 0.0
            c_Mk = 0.0
            c_REk = 0.0
            c_Sek = 0.0
            c_Ck = 0.0
            c_Enk = 0.0
            C_Ck = 0.0
            u_k = 1.0

            CostCenter.__init__(self, T_Nk, T_Pk, c_Lk, c_Eqk, c_Mk, c_REk, c_Sek, c_Ck, c_Enk, C_Ck, u_k)


class PartAssemblyCost(CostCenter):

    def __init__(self, cn, sd):

        d_w = sd.beam.d_w                       # Length of staigth portion of the beam web [mm]
        P = sd.beam.P                           # Perimeter of beam cross-section [mm]
        a_web = sd.weld_w                       # [mm]
        a_flange = sd.weld_f                    # [mm]
        cost_function = cn.cost_function

        if cost_function == 'Diaz':
            # Manufacturing cost for welding the end-plate
            L_f = (2.0*sd.beam.b + 2.0*(sd.beam.b - 2.0*sd.beam.r - sd.beam.t_w))*1.0e-3      # Flange weld length [m]
            L_w = (2.0*(sd.beam.h - 2.0*sd.beam.r - 2.0*sd.beam.t_f))*1.0e-3                  # Web weld length [m]
            T_w_af = 9.03*(a_flange*0.1)**2.0 + 4.68*(a_flange*0.1) - 0.82    # Flange welding time [min/m]
            T_w_aw = 9.03*(a_web*0.1)**2.0 + 4.68*(a_web*0.1) - 0.82          # Web welding time [min/m]
            k_w = Diaz['k_w']                                                 # Unit cost for welding [e/min]
            f_w = Diaz['f_w']                                                 # Welding factor that increases labor [-]
            T_w_ex = Diaz['T_w_ex']                                           # Additional welding time [min]
            welding_cost = k_w*(f_w*(T_w_af*L_f + T_w_aw*L_w) + T_w_ex)       # Total cost of welding operation

            # Cost of the welding consumables
            M_w_mc_af = 0.66*(a_flange*0.1)**2.0 + 0.18*(a_flange*0.1) - 0.10   # Material cost of welding flange
            M_w_mc_aw = 0.66*(a_web*0.1)**2.0 + 0.18*(a_web*0.1) - 0.10         # Material cost of welding web
            k_w_mt = Diaz['k_w_mt']                                             # Unit cost of weld [e/kg]
            welding_materials = k_w_mt*(M_w_mc_af*L_f + M_w_mc_aw*L_w)          # Total cost of welding materials

            self.cost = welding_cost + welding_materials        # Total cost of welding
        else:
            L = 15.0            # Workspace length [m]
            B = 5.0             # Workspace width [m]
            Area = L * B        # Workspace area [m^2]

            # Non-productive time
            T_Nk = 0.0

            # Productive time
            T_PTa = 1.59                    # Tackwelding time [min]
            L_fw_web = 2.0*d_w              # Flange weld length [mm]
            L_fw_flange = P - 2.0*d_w       # Web weld length [mm]
            # Welding time of web weld [min]
            T_Pfw_web = (L_fw_web/1000.0)*(0.4988*a_web**2.0 - 0.0005*a_web + 0.0021)
            # Welding time of flange weld [min]
            T_Pfw_flange = (L_fw_flange/1000.0)*(0.4988*a_flange**2.0 - 0.0005*a_flange + 0.0021)
            T_Pk = T_PTa + T_Pfw_web + T_Pfw_flange         # Total processing time [min]

            # Labour
            c_hour = 27.42              # Hourly cost [e/h]
            n_worker = 1.0              # Number of workers [-]
            c_min = c_hour / 60.0       # Minute rate of worker [e/min]
            c_Lk = n_worker * c_min     # Unit labor cost [e/min]

            # Equipment installment cost [e/min]
            P_invest_Eq = 5000.0        # Investment of equipment [e]
            P_resale_Eq = 0.0           # Resale value [e]
            I_Eq = 0.05                 # Interest rate (cost of capital) [%]
            n_Eq = 10.0                 # n = time [years]
            c_Eqk = annual_cost(P_invest_Eq - P_resale_Eq, I_Eq, n_Eq) / work_mins_in_a

            # Maintenance
            c_maintenance_Eq = 1000.0                       # Maintenance cost of equipment [e/a]
            c_Mk = c_maintenance_Eq / work_mins_in_a        # Maintenance cost of equipment [e/min]

            # Real estate installment cost [e/min]
            P_invest_RealS = Area * c_construction          # Investment cost of workspace area [e]
            P_resale_RealS = Area * c_land                  # Resale value of workspace [e]
            I_RealS = 0.05                                  # Interest rate (cost of capital) [%]
            n_RealS = 50.0                                  # n = time (years)
            c_REk = annual_cost(P_invest_RealS - P_resale_RealS, I_RealS, n_RealS) / work_mins_in_a

            # Real estate maintenance cost
            c_maintenance_RealS = Area * RE_maintenance     # Real estate maintenance cost[e]
            c_Sek = c_maintenance_RealS / work_mins_in_a    # Real estate maintenance cost[e/min]

            # Total cost of non-time-related consumables [e]
            # Consumables: welding wire, shield gas
            c_Ck = 0.0                                                          # Total cost of time related consumables
            C_Ck_web = L_fw_web*a_web**2.0*0.00000785*(1.91 + 4.44)             # Consumables cost for web fillet weld
            C_Ck_flange = L_fw_flange*a_flange**2.0*0.00000785*(1.91 + 4.44)    # Consumables cost for web fillet weld
            C_Ck = C_Ck_web + C_Ck_flange

            # Energy
            P_total = 6.0                               # Power of welding equipment
            c_Enk = P_total * (c_En_unit / 60.0)        # Cost of energy [e/min]

            u_k = 1.0       # Workspace utilization ratio, set to 1

            CostCenter.__init__(self, T_Nk, T_Pk, c_Lk, c_Eqk, c_Mk, c_REk, c_Sek, c_Ck, c_Enk, C_Ck, u_k)


class PostTreatmentCost(CostCenter):

    def __init__(self, cn, sd):

        weld_test_length = sd.beam.P
        test_method = cn.test_method
        P = weld_test_length                        # Perimeter of beam cross-section [mm]
        cost_function = cn.cost_function

        if cost_function == 'Diaz':
            self.cost = 0.0
        else:
            L = 35.0            # Workspace length [m]
            B = 10.0            # Workspace width [m]
            Area = L * B        # Workspace area [m^2]

            # Non-productive time [min]
            T_Nk = 0.0

            # Productive time, P=testing length of welds [mm]
            if test_method == "VT":       # Visual testing
                T_Pk = 0.0*P    # [min]
            elif test_method == "UT":     # Ultrasonic testing
                T_Pk = 0.06*P   # [min]
            elif test_method == "MT":     # Magnetic particle testing
                T_Pk = 0.015*P  # [min]
            else:
                print("Defined testing method \"{}\" not implemented. Used testing time 0.0 min".format(test_method))
                T_Pk = 0.0

            # Labour
            c_hour = 27.32              # Hourly cost [e/h]
            n_worker = 1.0              # Number of workers [-]
            c_min = c_hour / 60.0       # Minute rate of worker [e/min]
            c_Lk = n_worker * c_min     # Unit labor cost [e/min]

            c_Eqk = 0.0         # Equipment installment cost [e/min]
            c_Mk = 0.0          # Maintenance cost of equipment [e/min]

            # Real estate installment cost [e/min]
            P_invest_RealS = Area * c_construction          # Investment cost of workspace area [e]
            P_resale_RealS = Area * c_land                  # Resale value of workspace [e]
            I_RealS = 0.05                                  # Interest rate (cost of capital) [%]
            n_RealS = 50.0                                  # n = time (years)
            c_REk = annual_cost(P_invest_RealS - P_resale_RealS, I_RealS, n_RealS) / work_mins_in_a

            # Real estate maintenance cost
            c_maintenance_RealS = Area * RE_maintenance     # Real estate maintenance cost[e]
            c_Sek = c_maintenance_RealS / work_mins_in_a    # Real estate maintenance cost[e/min]

            c_Ck = 0.0          # Total cost of time related consumables
            C_Ck = 0.0          # Total cost of non-time-related consumables
            c_Enk = 0.0         # Total energy cost

            u_k = 1.0           # Workspace utilization ratio, set to 1

            CostCenter.__init__(self, T_Nk, T_Pk, c_Lk, c_Eqk, c_Mk, c_REk, c_Sek, c_Ck, c_Enk, C_Ck, u_k)


class PaintingCost(CostCenter):
    NOT_IMPLEMENTED = "NOT IMPLEMENTED YET! Add this feature: "

    def __init__(self, cn, sd):

        paint = cn.paint
        cost_function = cn.cost_function

        if cost_function == 'Diaz':
            A_p = 2.0*(sd.plate_width*sd.plate_height)      # Painting area [mm^2]

            k_p = Diaz['k_p']           # Unit cost for painting [e/min]
            T_p = Diaz['T_p']           # Time consumption of painting [min/m^2]
            k_p_mt = Diaz['k_p_mt']     # Unit cost of paint [e/l]
            M_p = Diaz['M_p']           # Paint consumption [l/m^2]
            self.cost = A_p*1.0e-6*(k_p*T_p + k_p_mt*M_p)      # Total cost of painting [e]
        else:
            L = 15.0        # Workspace length [m]
            B = 5.0         # Workspace width [m]
            Area = L * B    # Workspace area [m^2]

            # Non-productive time [min], drying
            T_Nk = 0.0

            # Painting area [mm], plate sides and beam cross-section taken into account
            A_p = 2.0*sd.plate_width*sd.plate_height + 2.0*sd.plate_thickness*(sd.plate_width + sd.plate_height) - \
                  sd.beam.A

            # Productive time [min], painting
            if paint == "alkyd":
                T_Pk = (0.513/900000)*A_p
            elif paint == "epoxy":
                T_Pk = (0.481/900000)*A_p
            elif paint == "polyurethane":
                T_Pk = (0.481/900000)*A_p
            elif paint == "acryl":
                T_Pk = (0.481/900000)*A_p
            else:
                T_Pk = 0.0
                print("Values for given paint \"{0}\" not implemented!".format(paint))

            # Labour
            c_hour = 27.59              # Hourly cost [e/h]
            n_worker = 1.0              # Number of workers [-]
            c_min = c_hour / 60.0       # Minute rate of worker [e/min]
            c_Lk = n_worker * c_min     # Unit labor cost [e/min]

            c_Eqk = 0.0     # Equipment installment cost [e/min]
            c_Mk = 0.0      # Maintenance cost of equipment [e/min]

            # Real estate installment cost [e/min]
            P_invest_RealS = Area * c_construction      # Investment cost of workspace area [e]
            P_resale_RealS = Area * c_land              # Resale value of workspace [e]
            I_RealS = 0.05                              # Interest rate (cost of capital) [%]
            n_RealS = 50.0                              # n = time (years)
            c_REk = annual_cost(P_invest_RealS - P_resale_RealS, I_RealS, n_RealS) / work_mins_in_a

            # Real estate maintenance cost
            c_maintenance_RealS = Area * RE_maintenance     # Real estate maintenance cost[e]
            c_Sek = c_maintenance_RealS / work_mins_in_a    # Real estate maintenance cost[e/min]

            # Consumables: Paint
            c_Ck = 0.0      # Total cost of time related consumables

            # Consumables: Paint
            # Total cost of non-time-related consumables [e]
            if paint == "alkyd":
                C_Ck = 3.87e-6*A_p
            elif paint == "epoxy":
                C_Ck = 4.11e-6*A_p
            elif paint == "polyurethane":
                C_Ck = 4.58e-6*A_p
            elif paint == "acryl":
                C_Ck = 5.22e-6*A_p
            else:
                C_Ck = 0.0
                print("Values for given paint \"{0}\" not implemented!".format(paint))

            c_Enk = 0.0     # Total energy cost

            u_k = 1.0       # Workspace utilization ratio, set to 1

            CostCenter.__init__(self, T_Nk, T_Pk, c_Lk, c_Eqk, c_Mk, c_REk, c_Sek, c_Ck, c_Enk, C_Ck, u_k)



# -*- coding: utf-8 -*-
# Copyright 2022 Kristo Mela
# This source code is licensed under the MIT license. See LICENSE in the repository root directory.
# Author(s): Kristo Mela
"""
Created on Wed May 15 22:17 2019

WQ sections (welded box beam)

@author: kmela
"""

from metku.eurocodes.en1993 import en1993_1_1, en1993_1_5, constants
from metku.sections.steel.steel_section import SteelSection

import matplotlib.pyplot as plt
import matplotlib.patches as patches

class WQSection(SteelSection):
    """ WQ-sections """

    def __init__(self, h, tw, b, tf, fy=355):
        """
            h -- height from top of profile to top of bottom flange
            b -- b[0] = width of top flange (between webs)
                 b[1] = width of bottom flange
            tw .. thickness of web
            tf .. tf[0] thickness of top flange
                  tf[1] thickness of bottom flange
            fy -- yield strength
        """
        self.h = h
        self.b = b
        self.tf = tf
        self.tw = tw
        
        #self.fy = fy
        
        A = self.area()
        Au = self.paint_area()
        I = self.second_moment()
        #Ashear = self.shear_area()
        Wel = self.section_modulus()
        Wpl = self.plastic_section_modulus()        
        
        
        super().__init__(fy, A, I, Au, Wpl, Wel, Ashear=1.0)

        self.Ashear = self.shear_area()

        """ Determine buckling curve: EN 1993-1-1, Table 6.2 """        
        self.imp_factor = [en1993_1_1.buckling_curve["b"],
                           en1993_1_1.buckling_curve["b"]]
        
    @property
    def height(self):
        """ Total height of the profile """
        return self.h+self.tb
    
    @property
    def hw(self):
        """ Part of the web between flanges"""
        return self.h - self.tf[0]
    @property
    def Aw(self):
        """ Area of one web plate """
        return self.h*self.tw
    
    @property
    def zw(self):
        """ Distance of centroid of web from bottom of section """
        return self.tb + 0.5*self.h

    @property
    def tb(self):
        """ Thickness of bottom flange """
        return self.tf[1]
    
    @property
    def bb(self):
        """ Width of bottom flange """
        return self.b[1]
    
    @property
    def Ab(self):
        """ Area of bottom flange """
        return self.bb*self.tb
    
    @property
    def zb(self):
        """ Distance of centroid of bottom flange from bottom of section """
        return 0.5*self.tb
    
    @property
    def tt(self):
        """ Thickness of top flange """
        return self.tf[0]
    
    @property
    def bt(self):
        """ Width of top flange """
        return self.b[0]
    
    @property
    def At(self):
        """ Area of top flange """
        return self.bt*self.tt
    
    @property
    def zt(self):
        """ Distance of centroid of top flange from bottom of section """
        return self.tb + self.h -0.5*self.tt
    
    @property
    def cf(self):
        """ Straight part of the flange """
        return self.b[0]

    @property
    def lip(self):
        """ length of outstand part of bottom flange (lip) """
        return 0.5*(self.bb-self.bt) - self.tw

    def flange_class(self,verb=False):
        """ Determine class of compressed flange """
        # cf = 0.5*(self.b - self.tw) - self.r
        rf = self.bt / self.tt
        cFlange = en1993_1_1.internal_part_in_compression(rf, self.eps)

        if verb:
            print("Flange classification (internal element):")
            print("cf = {0:4.2f}, tf = {1:4.2f}".format(self.bt,self.tt))
            print("cf/tf = {0:4.2f}".format(rf))
            print("Flange class = {0}".format(cFlange))

        return cFlange

    def web_class_comp(self,verb=False):
        """ Determine class of compressed web """
        rw = self.hw / self.tw
        cWeb = en1993_1_1.internal_part_in_compression(rw, self.eps)

        if verb:
            print("Web classification (internal part in compression):")
            print("hw = {0:4.2f}, tw = {1:4.2f}".format(self.hw,self.tw))
            print("hw/tw = {0:4.2f}".format(rw))
            print("Web class = {0}".format(cWeb))

        return cWeb

    def web_class_bend(self,verb=False):
        """ Determine class of web in bending """
        zel = self.elastic_neutral_axis()
        dpl = self.plastic_neutral_axis()
        rw = self.hw / self.tw
        psi = zel/(self.height-zel)
        alpha = dpl/self.hw
        cWeb = en1993_1_1.internal_part_comp_bend(rw, self.eps,alpha,psi)

        if verb:
            print("Web classification (internal part in compression and bending):")
            print("hw = {0:4.2f}, tw = {1:4.2f}".format(self.hw,self.tw))
            print("hw/tw = {0:4.2f}".format(rw))
            print("alpha = {0:4.2f}".format(alpha))
            print("psi = {0:4.2f}".format(psi))
            print("Web class = {0}".format(cWeb))

        return cWeb

    def web_class_comp_bend(self, Ned, verb=False):
        """ Determine class of web in combined bending and compression 
            TODO
        """
        # cw = self.h-2*self.tf-2*self.r
        rw = self.hw / self.tw
        Ac = 0.5 * (self.A + Ned / (self.fy / constants.gammaM0))
        a = Ac / self.A
        p = -1

        cWeb = en1993_1_1.internal_part_comp_bend(rw, self.eps, a, p)

        if verb:
            print("Web classification (internal part in compression and bending):")
            print("hw = {0:4.2f}, tw = {1:4.2f}".format(self.hw,self.tw))
            print("hw/tw = {0:4.2f}".format(rw))
            print("Web class = {0}".format(cWeb))

        return cWeb

    def moment_axial_force_interact(self, UN, MRd):
        """ Interaction rule for section resistance for combined
            axial force and bending
            
            input: UN .. NEd/NRd
                  MRd .. moment resistance           
            TODO
        """
        aw = min((self.A - 2 * self.b * self.tf) / self.A, 0.5)

        if UN > 1.0:
            MNRd = 0.0
        else:
            MNRd = MRd * min((1 - UN) / (1 - 0.5 * aw), 1)

        return MNRd

    def area(self):
        """ Cross-sectional area """
        return 2*self.Aw + self.At + self.Ab
    
    def shear_area(self):
        """ Shear area """
        return 2*en1993_1_5.shear_eta(self.fy)*self.hw*self.tw
    
    def paint_area(self):
        """ Area to be painted (circumference of the section) """
        return self.bb+2*self.tb + 2*self.lip + 2*(self.h+self.tw) + self.bt
        
    
    def elastic_neutral_axis(self):
        """ Elastic neutral axis 
            measured from the bottom of the section
        """
        zel = (self.zb*self.Ab + self.zw*self.Aw + self.zt*self.At)/self.area()
        return zel
    
    def plastic_neutral_axis(self):
        """ Plastic neutral axis
            measured downwards from the bottom of the top flange
        """
        # with respect to y-axis
        dpl = 0.5*(self.hw-self.tt)-0.25*(self.At-self.Ab)/self.tw
        return dpl
    
    def second_moment(self):
        """ Second moment of area with respect to both principal axes"""
        zel = self.elastic_neutral_axis()
                
        # with respect to z axis
        Iy = 1/12*self.bb*self.tb**3 + (zel-self.zb)**2*self.Ab + \
                1/6*self.tw*self.h**3 + (zel-self.zw)**2*2*self.Aw + \
                1/12*self.bt*self.tt**3 + (zel-self.zt)**2*self.At

        # with respect to z axis
        # yw .. distance of web centroid from z axis
        yw = 0.5*(self.bt+self.tw)
        Iz = 1/12*self.tb*self.bb**3 + 1/12*self.tt*self.bt**3 + \
                1/6*self.h*self.tw**3 + yw**2*2*self.Aw

        return [Iy,Iz]
    
    def section_modulus(self):
        """ Elastic section modulus for bending """
        
        I = self.second_moment()
        zel = self.elastic_neutral_axis()
        
        # distance of neutral axis from the top of the section
        ztop = self.h+self.bb - zel
        
        Wel =[0,0]
        Wel[0] = I[0]/max(zel,ztop)
        Wel[1] = I[1]/(0.5*self.bb)
        
        return Wel
    
    def plastic_section_modulus(self):
        """ Plastic section modulus for bending """
        
        """ distance of plastic neutral axis from the bottom of the top flange """
        Wpl = [0,0]
    
        # with respect to y-axis
        dpl = self.plastic_neutral_axis()

        
        if dpl < 0:
            # neutral axis is in the top flange
            # TODO
            Wpl[0] = 0
        elif dpl > self.hw:
            # neutral axis is in the bottom flange
            # TODO
            Wpl[0] = 0        
        else:
            # neutral axis is in the web
            rt = dpl + 0.5*self.tt
            rw1 = 0.5*(dpl+self.tt)
            rw2 = 0.5*(self.hw-dpl)
            rb = self.hw-dpl+0.5*self.tb
            
            Wpl[0] = self.At*rt + 2*(dpl+self.tt)*self.tw*rw1 + \
                    2*(self.hw-dpl)*self.tw*rw2 + self.Ab*rb
            
        # with respect to z-axis
        Wpl[1] = 1/4*self.Ab*self.bb + 1/4*self.At*self.bt + \
                    self.Aw*(self.bt+self.tw)
        return Wpl
    
    def draw(self,axes=None,origin=[0,0],plot_centroid=False,
             plot_plastic_centroid=False):
        """ Draw the profile 
            input:
                axes .. matplotlib.axes.Axes object. If this is given,
                        the profile will be drawn to that Axes
                theta .. rotation (optional)
        """
        
        if axes is None:
            fig, ax = plt.subplots(1)
        else:
            ax = axes
        
        # create section parts
        bot_flange = patches.Rectangle((0,0),width=self.bb,height=self.tb,\
                                       fill=False, hatch='\\')
        web1 = patches.Rectangle((self.lip,self.tb),width=self.tw,height=self.h,\
                                 fill=False, hatch='\\')
        web2 = patches.Rectangle((self.bb-self.lip-self.tw,self.tb),\
                                 width=self.tw,height=self.h,\
                                 fill=False, hatch='\\')
        top_flange = patches.Rectangle((self.lip+self.tw,self.hw+self.tb),\
                                       width=self.bt,height=self.tt,\
                                       fill=False, hatch='\\')
        ax.add_patch(bot_flange)
        ax.add_patch(web1)
        ax.add_patch(web2)
        ax.add_patch(top_flange)
        ax.set_xlim((0,self.bb))
        ax.set_ylim((0,self.tb+self.h))
        
        if plot_centroid:
            zel = self.elastic_neutral_axis()
            ax.plot(0.5*self.bb,zel,'or')
        
        if plot_plastic_centroid:
            dpl = self.plastic_neutral_axis()        
        
            ax.plot(0.5*self.bb,self.tb+self.hw-dpl,'db')
        
        plt.show()
        # bottom flange coordinates
        #bf_xy = [[0,0],[self.bb,0.0],[self.bb,self.tb],[0,self.tb]]
        
        return ax
        

if __name__ == '__main__':
    p = WQSection(390, 10, [200,400], [20,25])
    #print(p.elastic_neutral_axis())
    p.web_class_bend(verb=True)
    p.flange_class(verb=True)
    p.draw()


# -*- coding: utf-8 -*-
"""
Created on Wed Mar 14 19:04:25 2018

Eurocodes package

@author: kmela
"""

from . import ISection
from .steel_section import *
from .ISection import *
from .RHS import *
from .CHS import *
from .catalogue import *
from .wi import *



# -*- coding: utf-8 -*-
# Copyright 2022 Kristo Mela
# This source code is licensed under the MIT license. See LICENSE in the repository root directory.
# Author(s): Kristo Mela

from metku.frame2d.frame2d import *
from metku.optimization.structopt import *

ten_bar_DLM = [41, 0, 38, 31, 0, 0, 27, 38, 37, 0]
ten_bar_TwoPhase = [41, 0, 38, 32, 0, 0, 27, 37, 37, 0]
ten_bar_PSO = [40, 0, 38, 29, 0, 0, 27, 39, 37, 1]
# [21607.5, 1044.9, 14770.5, 9997.5, 1044.9, 1044.9, 5140.65, 14190.0, 14190.0, 1044.9]


TEN_BAR_AREAS_in2 = [1.62, 1.80, 1.99, 2.13, 2.38, 2.62, 2.63, 2.88, 2.93,
                     3.09, 3.13, 3.38, 3.47, 3.55, 3.63, 3.84, 3.87, 3.88,
                     4.18, 4.22, 4.49, 4.59, 4.80, 4.97, 5.12, 5.74, 7.22,
                     7.97, 11.50, 13.50, 13.90, 14.20, 15.50, 16.00, 16.90,
                     18.80, 19.90, 22.00, 22.90, 26.50, 30.00, 33.50]

TEN_BAR_AREAS_mm2 = [1044.9, 1161.0, 1283.55, 1373.85, 1535.1, 1689.9,
                     1696.35, 1857.6, 1889.85, 1993.05, 2018.85, 2180.1,
                     2238.15, 2289.75, 2341.35, 2476.8, 2496.15, 2502.6,
                     2696.1, 2721.9, 2896.05, 2960.55, 3096.0, 3205.65,
                     3302.4, 3702.3, 4656.9, 5140.65, 7417.5, 8707.5, 8965.5,
                     9159.0, 9997.5, 10320.0, 10900.5, 12126.0, 12835.5,
                     14190.0, 14770.5, 17092.5, 19350.0, 21607.5]


class TenBarTruss(OptimizationProblem):
    # Problem parameters
    L = 9144  # mm
    F = 444_890  # N
    E = 68950  # MPa
    rho = 2768e-9  # kg/mm3

    properties = {
        'L': str(L) + " mm",
        'F': f'{F / 1e3 :.2f} kN',
        'E': f'{E / 1e3 :.2f} MPa',
        'rho': f'{rho * 1e9 :.0f} kg/m3'
    }

    # Constraint limits
    sigma_max = 172.37  # MPa
    delta_max = 50.8  # mm

    def __init__(self, prob_type="discrete"):

        super().__init__(name='TenBarTruss')
        self.prob_type = prob_type
        self.structure = self._create_structure()
        self.create_variables(profiles=TEN_BAR_AREAS_mm2)
        self.create_constraints()
        self.create_objective()

    def create_variables(self, profiles=[0, 1e6]):
        """
        Creates variables used in optimization

        Appends each variable to 'vars' 'list from where they can be accessed
        """
        self.vars = []
        for i, mem in enumerate(self.structure.members.values()):
            name = 'A' + str(i + 1)
            if self.prob_type == "discrete":
                var = DiscreteVariable(name,
                                       values=profiles,
                                       target={"property": "A",
                                               "objects": [mem]})

            elif self.prob_type == "continuous":
                var = Variable(name,
                               lb=profiles[0],
                               ub=profiles[-1],
                               target={"property": "A",
                                       "objects": [mem]})

            elif self.prob_type == 'index':
                var = IndexVariable(name,
                                    values=profiles,
                                    target={"property": "A",
                                            "objects": [mem]})





            else:

                raise TypeError("Problem type must be either 'discrete' "
                                "or 'continuous")

            self.add(var)

    def _create_structure(self):
        """
        Creates the ten bar truss structure

        :return: structure
        :rtype: Frame2D
        """

        frame = Frame2D(num_elements=1)
        # Members
        n1 = [2 * self.L, self.L]
        n2 = [2 * self.L, 0]
        n3 = [self.L, self.L]
        n4 = [self.L, 0]
        n5 = [0, 0]
        n6 = [0, self.L]
        bar1 = SteelBeam([n6, n3])
        bar2 = SteelBeam([n3, n1])
        bar3 = SteelBeam([n5, n4])
        bar4 = SteelBeam([n4, n2])
        bar5 = SteelBeam([n3, n4])
        bar6 = SteelBeam([n1, n2])
        bar7 = SteelBeam([n6, n4])
        bar8 = SteelBeam([n5, n3])
        bar9 = SteelBeam([n3, n2])
        bar10 = SteelBeam([n4, n1])
        frame.add(bar1)
        frame.add(bar2)
        frame.add(bar3)
        frame.add(bar4)
        frame.add(bar5)
        frame.add(bar6)
        frame.add(bar7)
        frame.add(bar8)
        frame.add(bar9)
        frame.add(bar10)
        # Supports
        frame.add(XYHingedSupport(n5))
        frame.add(XYHingedSupport(n6))
        # Point loads
        frame.add(PointLoad(n4, [0, -self.F, 0], load_id=LoadIDs.ULS))
        frame.add(PointLoad(n2, [0, -self.F, 0], load_id=LoadIDs.ULS))
        # Change material properties
        for mem in frame.members.values():
            mem.material.fy = self.sigma_max
            mem.E = self.E  # MPa
            mem.A = TEN_BAR_AREAS_mm2[0]
            mem.rho = self.rho
            mem.Sj1 = 0
            mem.Sj2 = 0

        frame.generate()
        frame.calculate()

        return frame

    def constraint_generator(self, mem):

        def compression_fun(x):
            self.substitute_variables(x)
            self.structure.calculate()
            return -mem.NEd[LoadIDs.ULS] / (mem.A * mem.fy) - 1

        def tension_fun(x):
            self.substitute_variables(x)
            self.structure.calculate()
            return mem.NEd[LoadIDs.ULS] / (mem.A * mem.fy) - 1

        def disp_fun(x):
            self.substitute_variables(x)
            self.structure.calculate()
            displacements = mem.nodal_displacements[LoadIDs.ULS]
            # disp = [dx, dy, rz]
            max_vals = [disp[1] for disp in displacements.values()]
            min_vals = [disp[1] for disp in displacements.values()]

            max_val = max(max_vals)
            min_val = min(min_vals)

            abs_max = max(max_val, abs(min_val))
            return abs_max / self.delta_max - 1

        return compression_fun, tension_fun, disp_fun

    def create_constraints(self):

        # Initialize constraints as an empty list
        self.cons = []
        i = 0
        for var in self.vars:
            for mem in var.target["objects"]:
                if isinstance(mem, FrameMember):
                    i += 1
                    compression_fun, tension_fun, disp_fun = self.constraint_generator(mem)

                    comp_con = NonLinearConstraint(con_fun=compression_fun,
                                                   name="Compression " + str(i),
                                                   )
                    comp_con.fea_required = True

                    tension_con = NonLinearConstraint(con_fun=tension_fun,
                                                      name="Tension " + str(i))
                    tension_con.fea_required = True

                    disp_con = NonLinearConstraint(con_fun=disp_fun,
                                                   name='Displacement ' + str(i),
                                                   )
                    disp_con.fea_required = True

                    self.add(comp_con)
                    self.add(tension_con)
                    self.add(disp_con)

    def create_objective(self):

        def objective(X):
            self.substitute_variables(X)
            weight = 0
            for mem in self.structure.members.values():
                weight += mem.length * mem.A * self.rho
            return weight

        obj = ObjectiveFunction(name="Weight",
                                obj_fun=objective)

        self.add(obj)


if __name__ == '__main__':
    from metku.optimization.solvers import SLP, MMA, SLSQP


    problem = TenBarTruss('continuous')
    # solver = SLP([0.05, 0.05], 1e-2)
    x0 = [np.random.randint(var.lb, var.ub) for var in problem.vars]
    # x0 = [var.ub for var in problem.vars]
    solver = MMA()


    #problem(x0)
    fopt, xopt = solver.solve(problem, x0=x0, verb=True, maxiter=200)
    print(f"{x0=}")
    problem(xopt)
    #problem.structure.plot_deflection(20)

